# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import TypeVar
from abc import ABC, abstractmethod
import time

from attr import dataclass

from mautrix.types import EventID, RoomID


@dataclass
class AbstractDisappearingMessage(ABC):
    room_id: RoomID
    event_id: EventID
    expiration_seconds: int
    expiration_ts: int | None = None

    @abstractmethod
    async def insert(self) -> None:
        pass

    @abstractmethod
    async def update(self) -> None:
        pass

    def start_timer(self) -> None:
        self.expiration_ts = int(time.time() * 1000) + (self.expiration_seconds * 1000)

    @abstractmethod
    async def delete(self) -> None:
        pass

    @classmethod
    @abstractmethod
    async def get_all_scheduled(cls: type[DisappearingMessage]) -> list[DisappearingMessage]:
        pass

    @classmethod
    @abstractmethod
    async def get_unscheduled_for_room(
        cls: type[DisappearingMessage], room_id: RoomID
    ) -> list[DisappearingMessage]:
        pass


DisappearingMessage = TypeVar("DisappearingMessage", bound=AbstractDisappearingMessage)



# Copyright (c) 2023 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, List, Optional, Union
import asyncio
import uuid

from mautrix.types import (
    DecryptedOlmEvent,
    DeviceID,
    EncryptionAlgorithm,
    EventType,
    ForwardedRoomKeyEventContent,
    IdentityKey,
    KeyRequestAction,
    RequestedKeyInfo,
    RoomID,
    RoomKeyRequestEventContent,
    SessionID,
    UserID,
)

from .base import BaseOlmMachine
from .sessions import InboundGroupSession


class KeyRequestingMachine(BaseOlmMachine):
    async def request_room_key(
        self,
        room_id: RoomID,
        sender_key: IdentityKey,
        session_id: SessionID,
        from_devices: Dict[UserID, List[DeviceID]],
        timeout: Optional[Union[int, float]] = None,
    ) -> bool:
        """
        Request keys for a Megolm group session from other devices.

        Once the keys are received, or if this task is cancelled (via the ``timeout`` parameter),
        a cancel request event is sent to the remaining devices. If the ``timeout`` is set to zero
        or less, this will return immediately, and the extra key requests will not be cancelled.

        Args:
            room_id: The room where the session is used.
            sender_key: The key of the user who created the session.
            session_id: The ID of the session.
            from_devices: A dict from user ID to list of device IDs whom to ask for the keys.
            timeout: The maximum number of seconds to wait for the keys. If the timeout is
                     ``None``, the wait time is not limited, but the task can still be cancelled.
                     If it's zero or less, this returns immediately and will never cancel requests.

        Returns:
            ``True`` if the keys were received and are now in the crypto store,
            ``False`` otherwise (including if the method didn't wait at all).
        """
        request_id = str(uuid.uuid1())
        request = RoomKeyRequestEventContent(
            action=KeyRequestAction.REQUEST,
            body=RequestedKeyInfo(
                algorithm=EncryptionAlgorithm.MEGOLM_V1,
                room_id=room_id,
                sender_key=sender_key,
                session_id=session_id,
            ),
            request_id=request_id,
            requesting_device_id=self.client.device_id,
        )

        wait = timeout is None or timeout > 0
        fut: Optional[asyncio.Future] = None
        if wait:
            fut = asyncio.get_running_loop().create_future()
            self._key_request_waiters[session_id] = fut
        await self.client.send_to_device(
            EventType.ROOM_KEY_REQUEST,
            {
                user_id: {device_id: request for device_id in devices}
                for user_id, devices in from_devices.items()
            },
        )
        if not wait:
            # Timeout is set and <=0, don't wait for keys
            return False
        assert fut is not None
        got_keys = False
        try:
            user_id, device_id = await asyncio.wait_for(fut, timeout=timeout)
            got_keys = True
            try:
                del from_devices[user_id][device_id]
                if len(from_devices[user_id]) == 0:
                    del from_devices[user_id]
            except KeyError:
                pass
        except (asyncio.CancelledError, asyncio.TimeoutError):
            pass
        del self._key_request_waiters[session_id]
        if len(from_devices) > 0:
            cancel = RoomKeyRequestEventContent(
                action=KeyRequestAction.CANCEL,
                request_id=str(request_id),
                requesting_device_id=self.client.device_id,
            )
            await self.client.send_to_device(
                EventType.ROOM_KEY_REQUEST,
                {
                    user_id: {device_id: cancel for device_id in devices}
                    for user_id, devices in from_devices.items()
                },
            )
        return got_keys

    async def _receive_forwarded_room_key(self, evt: DecryptedOlmEvent) -> None:
        key: ForwardedRoomKeyEventContent = evt.content
        if await self.crypto_store.has_group_session(key.room_id, key.session_id):
            self.log.debug(
                f"Ignoring received session {key.session_id} from {evt.sender}/"
                f"{evt.sender_device}, as crypto store says we have it already"
            )
            return
        if not key.beeper_max_messages or not key.beeper_max_age_ms:
            await self._fill_encryption_info(key)
        key.forwarding_key_chain.append(evt.sender_key)
        sess = InboundGroupSession.import_session(
            key.session_key,
            key.signing_key,
            key.sender_key,
            key.room_id,
            key.forwarding_key_chain,
            max_age=key.beeper_max_age_ms,
            max_messages=key.beeper_max_messages,
            is_scheduled=key.beeper_is_scheduled,
        )
        if key.session_id != sess.id:
            self.log.warning(
                f"Mismatched session ID while importing forwarded key from "
                f"{evt.sender}/{evt.sender_device}: '{key.session_id}' != '{sess.id}'"
            )
            return

        await self.crypto_store.put_group_session(
            key.room_id, key.sender_key, key.session_id, sess
        )
        self._mark_session_received(key.session_id)

        self.log.debug(
            f"Imported {key.session_id} for {key.room_id} "
            f"from {evt.sender}/{evt.sender_device}"
        )

        try:
            task = self._key_request_waiters[key.session_id]
        except KeyError:
            pass
        else:
            task.set_result((evt.sender, evt.sender_device))



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Optional
import json

from ..primitive import JSON
from ..util import ExtensibleEnum, Serializable, SerializableEnum


class RoomType(ExtensibleEnum):
    SPACE = "m.space"


class EventType(Serializable):
    """
    An immutable enum-like class that represents a specific Matrix event type.

    In addition to the plain event type string, this also includes the context that the event is
    used in (see: :class:`Class`). Comparing ``EventType`` instances for equality will check both
    the type string and the class.

    The idea behind the wrapper is that incoming event parsers will always create an ``EventType``
    instance with the correct class, regardless of what the usual context for the event is. Then
    when the event is being handled, the type will not be equal to ``EventType`` instances with a
    different class. For example, if someone sends a non-state ``m.room.name`` event, checking
    ``if event.type == EventType.ROOM_NAME`` would return ``False``, because the class would be
    different. Bugs caused by not checking the context of an event (especially state event vs
    message event) were very common in the past, and using a wrapper like this helps prevent them.
    """

    _by_event_type = {}

    class Class(SerializableEnum):
        """The context that an event type is used in."""

        UNKNOWN = "unknown"

        STATE = "state"
        """Room state events"""

        MESSAGE = "message"
        """Room message events, i.e. room events that are not state events"""

        ACCOUNT_DATA = "account_data"
        """Account data events, user-specific storage used for synchronizing info between clients.
        Can be global or room-specific."""

        EPHEMERAL = "ephemeral"
        """Ephemeral events. Currently only typing notifications, read receipts and presence are
        in this class, as custom ephemeral events are not yet possible."""

        TO_DEVICE = "to_device"
        """Device-to-device events, primarily used for exchanging encryption keys"""

    __slots__ = ("t", "t_class")

    t: str
    """The type string of the event."""
    t_class: Class
    """The context where the event appeared."""

    def __init__(self, t: str, t_class: Class) -> None:
        object.__setattr__(self, "t", t)
        object.__setattr__(self, "t_class", t_class)
        if t not in self._by_event_type:
            self._by_event_type[t] = self

    def serialize(self) -> JSON:
        return self.t

    @classmethod
    def deserialize(cls, raw: JSON) -> Any:
        return cls.find(raw)

    @classmethod
    def find(cls, t: str, t_class: Optional[Class] = None) -> "EventType":
        """
        Create a new ``EventType`` instance with the given type and class.

        If an ``EventType`` instance with the same type string and class has been created before,
        or if no class is specified here, this will return the same instance instead of making a
        new one.

        Examples:
            >>> from mautrix.client import Client
            >>> from mautrix.types import EventType
            >>> MY_CUSTOM_TYPE = EventType.find("com.example.custom_event", EventType.Class.STATE)
            >>> client = Client(...)
            >>> @client.on(MY_CUSTOM_TYPE)
            ... async def handle_event(evt): ...

        Args:
            t: The type string.
            t_class: The class of the event type.

        Returns:
            An ``EventType`` instance with the given parameters.
        """
        try:
            return cls._by_event_type[t].with_class(t_class)
        except KeyError:
            return EventType(t, t_class=t_class or cls.Class.UNKNOWN)

    def json(self) -> str:
        return json.dumps(self.serialize())

    @classmethod
    def parse_json(cls, data: str) -> "EventType":
        return cls.deserialize(json.loads(data))

    def __setattr__(self, *args, **kwargs) -> None:
        raise TypeError("EventTypes are frozen")

    def __delattr__(self, *args, **kwargs) -> None:
        raise TypeError("EventTypes are frozen")

    def __str__(self):
        return self.t

    def __repr__(self):
        return f'EventType("{self.t}", EventType.Class.{self.t_class.name})'

    def __hash__(self):
        return hash(self.t) ^ hash(self.t_class)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, EventType):
            return False
        return self.t == other.t and self.t_class == other.t_class

    def with_class(self, t_class: Optional[Class]) -> "EventType":
        """Return a copy of this ``EventType`` with the given class. If the given class is the
        same as what this instance has, or if the given class is ``None``, this returns ``self``
        instead of making a copy."""
        if t_class is None or self.t_class == t_class:
            return self
        return EventType(t=self.t, t_class=t_class)

    @property
    def is_message(self) -> bool:
        """A shortcut for ``type.t_class == EventType.Class.MESSAGE``"""
        return self.t_class == EventType.Class.MESSAGE

    @property
    def is_state(self) -> bool:
        """A shortcut for ``type.t_class == EventType.Class.STATE``"""
        return self.t_class == EventType.Class.STATE

    @property
    def is_ephemeral(self) -> bool:
        """A shortcut for ``type.t_class == EventType.Class.EPHEMERAL``"""
        return self.t_class == EventType.Class.EPHEMERAL

    @property
    def is_account_data(self) -> bool:
        """A shortcut for ``type.t_class == EventType.Class.ACCOUNT_DATA``"""
        return self.t_class == EventType.Class.ACCOUNT_DATA

    @property
    def is_to_device(self) -> bool:
        """A shortcut for ``type.t_class == EventType.Class.TO_DEVICE``"""
        return self.t_class == EventType.Class.TO_DEVICE


_standard_types = {
    EventType.Class.STATE: {
        "m.room.canonical_alias": "ROOM_CANONICAL_ALIAS",
        "m.room.create": "ROOM_CREATE",
        "m.room.join_rules": "ROOM_JOIN_RULES",
        "m.room.member": "ROOM_MEMBER",
        "m.room.power_levels": "ROOM_POWER_LEVELS",
        "m.room.history_visibility": "ROOM_HISTORY_VISIBILITY",
        "m.room.name": "ROOM_NAME",
        "m.room.topic": "ROOM_TOPIC",
        "m.room.avatar": "ROOM_AVATAR",
        "m.room.pinned_events": "ROOM_PINNED_EVENTS",
        "m.room.tombstone": "ROOM_TOMBSTONE",
        "m.room.encryption": "ROOM_ENCRYPTION",
        "m.space.child": "SPACE_CHILD",
        "m.space.parent": "SPACE_PARENT",
    },
    EventType.Class.MESSAGE: {
        "m.room.redaction": "ROOM_REDACTION",
        "m.room.message": "ROOM_MESSAGE",
        "m.room.encrypted": "ROOM_ENCRYPTED",
        "m.sticker": "STICKER",
        "m.reaction": "REACTION",
        "m.call.invite": "CALL_INVITE",
        "m.call.candidates": "CALL_CANDIDATES",
        "m.call.select_answer": "CALL_SELECT_ANSWER",
        "m.call.answer": "CALL_ANSWER",
        "m.call.hangup": "CALL_HANGUP",
        "m.call.reject": "CALL_REJECT",
        "m.call.negotiate": "CALL_NEGOTIATE",
        "com.beeper.message_send_status": "BEEPER_MESSAGE_STATUS",
    },
    EventType.Class.EPHEMERAL: {
        "m.receipt": "RECEIPT",
        "m.typing": "TYPING",
        "m.presence": "PRESENCE",
    },
    EventType.Class.ACCOUNT_DATA: {
        "m.direct": "DIRECT",
        "m.push_rules": "PUSH_RULES",
        "m.tag": "TAG",
        "m.ignored_user_list": "IGNORED_USER_LIST",
    },
    EventType.Class.TO_DEVICE: {
        "m.room.encrypted": "TO_DEVICE_ENCRYPTED",
        "m.room_key": "ROOM_KEY",
        "m.room_key.withheld": "ROOM_KEY_WITHHELD",
        "org.matrix.room_key.withheld": "ORG_MATRIX_ROOM_KEY_WITHHELD",
        "m.room_key_request": "ROOM_KEY_REQUEST",
        "m.forwarded_room_key": "FORWARDED_ROOM_KEY",
        "m.dummy": "TO_DEVICE_DUMMY",
        "com.beeper.room_key.ack": "BEEPER_ROOM_KEY_ACK",
    },
    EventType.Class.UNKNOWN: {
        "__ALL__": "ALL",  # This is not a real event type
    },
}

for _t_class, _types in _standard_types.items():
    for _t, _name in _types.items():
        _event_type = EventType(t=_t, t_class=_t_class)
        setattr(EventType, _name, _event_type)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pytest

from .format_duration import format_duration

tests = {
    1234: "20 minutes and 34 seconds",
    987654: "1 week, 4 days, 10 hours, 20 minutes and 54 seconds",
    694861: "1 week, 1 day, 1 hour, 1 minute and 1 second",
    1: "1 second",
    59: "59 seconds",
    60: "1 minute",
    120: "2 minutes",
}


def test_format_duration() -> None:
    for seconds, formatted in tests.items():
        assert format_duration(seconds) == formatted


def test_non_positive_error() -> None:
    with pytest.raises(ValueError):
        format_duration(0)

    with pytest.raises(ValueError):
        format_duration(-123)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# Partly based on github.com/Cadair/python-appservice-framework (MIT license)
from __future__ import annotations

from typing import Any, Awaitable, Callable
from json import JSONDecodeError
import asyncio
import json
import logging

from aiohttp import web

from mautrix.types import (
    JSON,
    ASToDeviceEvent,
    DeviceID,
    DeviceLists,
    DeviceOTKCount,
    EphemeralEvent,
    Event,
    EventType,
    RoomAlias,
    SerializerError,
    UserID,
)
from mautrix.util import background_task

HandlerFunc = Callable[[Event], Awaitable]


class AppServiceServerMixin:
    loop: asyncio.AbstractEventLoop
    log: logging.Logger

    hs_token: str
    ephemeral_events: bool
    encryption_events: bool

    query_user: Callable[[UserID], JSON]
    query_alias: Callable[[RoomAlias], JSON]

    transactions: set[str]
    event_handlers: list[HandlerFunc]
    to_device_handler: HandlerFunc | None
    otk_handler: Callable[[dict[UserID, dict[DeviceID, DeviceOTKCount]]], Awaitable] | None
    device_list_handler: Callable[[DeviceLists], Awaitable] | None

    def __init__(self, ephemeral_events: bool = False, encryption_events: bool = False) -> None:
        self.transactions = set()
        self.event_handlers = []
        self.to_device_handler = None
        self.otk_handler = None
        self.device_list_handler = None
        self.ephemeral_events = ephemeral_events
        self.encryption_events = encryption_events

        async def default_query_handler(_):
            return None

        self.query_user = default_query_handler
        self.query_alias = default_query_handler

    def register_routes(self, app: web.Application) -> None:
        app.router.add_route(
            "PUT", "/transactions/{transaction_id}", self._http_handle_transaction
        )
        app.router.add_route("GET", "/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/users/{user_id}", self._http_query_user)
        app.router.add_route(
            "PUT", "/_matrix/app/v1/transactions/{transaction_id}", self._http_handle_transaction
        )
        app.router.add_route("GET", "/_matrix/app/v1/rooms/{alias}", self._http_query_alias)
        app.router.add_route("GET", "/_matrix/app/v1/users/{user_id}", self._http_query_user)
        app.router.add_route("POST", "/_matrix/app/v1/ping", self._http_ping)
        app.router.add_route("POST", "/_matrix/app/unstable/fi.mau.msc2659/ping", self._http_ping)

    def _check_token(self, request: web.Request) -> bool:
        try:
            token = request.rel_url.query["access_token"]
        except KeyError:
            try:
                token = request.headers["Authorization"].removeprefix("Bearer ")
            except KeyError:
                self.log.debug("No access_token nor Authorization header in request")
                return False

        if token != self.hs_token:
            self.log.debug(f"Incorrect hs_token in request")
            return False

        return True

    async def _http_query_user(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.json_response({"error": "Invalid auth token"}, status=401)

        try:
            user_id = request.match_info["user_id"]
        except KeyError:
            return web.json_response({"error": "Missing user_id parameter"}, status=400)

        try:
            response = await self.query_user(user_id)
        except Exception:
            self.log.exception("Exception in user query handler")
            return web.json_response({"error": "Internal appservice error"}, status=500)

        if not response:
            return web.json_response({}, status=404)
        return web.json_response(response)

    async def _http_query_alias(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            return web.json_response({"error": "Invalid auth token"}, status=401)

        try:
            alias = request.match_info["alias"]
        except KeyError:
            return web.json_response({"error": "Missing alias parameter"}, status=400)

        try:
            response = await self.query_alias(alias)
        except Exception:
            self.log.exception("Exception in alias query handler")
            return web.json_response({"error": "Internal appservice error"}, status=500)

        if not response:
            return web.json_response({}, status=404)
        return web.json_response(response)

    async def _http_ping(self, request: web.Request) -> web.Response:
        if not self._check_token(request):
            raise web.HTTPUnauthorized(
                content_type="application/json",
                text=json.dumps({"error": "Invalid auth token", "errcode": "M_UNKNOWN_TOKEN"}),
            )
        try:
            body = await request.json()
        except JSONDecodeError:
            raise web.HTTPBadRequest(
                content_type="application/json",
                text=json.dumps({"error": "Body is not JSON", "errcode": "M_NOT_JSON"}),
            )
        txn_id = body.get("transaction_id")
        self.log.info(f"Received ping from homeserver with transaction ID {txn_id}")
        return web.json_response({})

    @staticmethod
    def _get_with_fallback(
        json: dict[str, Any], field: str, unstable_prefix: str, default: Any = None
    ) -> Any:
        try:
            return json.pop(field)
        except KeyError:
            try:
                return json.pop(f"{unstable_prefix}.{field}")
            except KeyError:
                return default

    async def _read_transaction_header(self, request: web.Request) -> tuple[str, dict[str, Any]]:
        if not self._check_token(request):
            raise web.HTTPUnauthorized(
                content_type="application/json",
                text=json.dumps({"error": "Invalid auth token", "errcode": "M_UNKNOWN_TOKEN"}),
            )

        transaction_id = request.match_info["transaction_id"]
        if transaction_id in self.transactions:
            raise web.HTTPOk(content_type="application/json", text="{}")

        try:
            return transaction_id, await request.json()
        except JSONDecodeError:
            raise web.HTTPBadRequest(
                content_type="application/json",
                text=json.dumps({"error": "Body is not JSON", "errcode": "M_NOT_JSON"}),
            )

    async def _http_handle_transaction(self, request: web.Request) -> web.Response:
        transaction_id, data = await self._read_transaction_header(request)

        txn_content_log = []
        try:
            events = data.pop("events")
            if events:
                txn_content_log.append(f"{len(events)} PDUs")
        except KeyError:
            raise web.HTTPBadRequest(
                content_type="application/json",
                text=json.dumps(
                    {"error": "Missing events object in body", "errcode": "M_BAD_JSON"}
                ),
            )

        if self.ephemeral_events:
            ephemeral = self._get_with_fallback(data, "ephemeral", "de.sorunome.msc2409")
            if ephemeral:
                txn_content_log.append(f"{len(ephemeral)} EDUs")
        else:
            ephemeral = None
        if self.encryption_events:
            to_device = self._get_with_fallback(data, "to_device", "de.sorunome.msc2409")
            device_lists = DeviceLists.deserialize(
                self._get_with_fallback(data, "device_lists", "org.matrix.msc3202")
            )
            otk_counts = {
                user_id: {
                    device_id: DeviceOTKCount.deserialize(count)
                    for device_id, count in devices.items()
                }
                for user_id, devices in self._get_with_fallback(
                    data, "device_one_time_keys_count", "org.matrix.msc3202", default={}
                ).items()
            }
            if to_device:
                txn_content_log.append(f"{len(to_device)} to-device events")
            if device_lists.changed:
                txn_content_log.append(f"{len(device_lists.changed)} device list changes")
            if otk_counts:
                txn_content_log.append(
                    f"{sum(len(vals) for vals in otk_counts.values())} OTK counts"
                )
        else:
            otk_counts = {}
            device_lists = None
            to_device = None

        if len(txn_content_log) > 2:
            txn_content_log = [", ".join(txn_content_log[:-1]), txn_content_log[-1]]
        if not txn_content_log:
            txn_description = "nothing?"
        else:
            txn_description = " and ".join(txn_content_log)
        self.log.debug(f"Handling transaction {transaction_id} with {txn_description}")

        try:
            output = await self.handle_transaction(
                transaction_id,
                events=events,
                extra_data=data,
                ephemeral=ephemeral,
                to_device=to_device,
                device_lists=device_lists,
                otk_counts=otk_counts,
            )
        except Exception:
            self.log.exception("Exception in transaction handler")
            output = None
        finally:
            self.log.debug(f"Finished handling transaction {transaction_id}")

        self.transactions.add(transaction_id)

        return web.json_response(output or {})

    @staticmethod
    def _fix_prev_content(raw_event: JSON) -> None:
        try:
            if raw_event["unsigned"] is None:
                del raw_event["unsigned"]
        except KeyError:
            pass
        try:
            raw_event["unsigned"]["prev_content"]
        except KeyError:
            try:
                raw_event.setdefault("unsigned", {})["prev_content"] = raw_event["prev_content"]
            except KeyError:
                pass

    async def handle_transaction(
        self,
        txn_id: str,
        *,
        events: list[JSON],
        extra_data: JSON,
        ephemeral: list[JSON] | None = None,
        to_device: list[JSON] | None = None,
        otk_counts: dict[UserID, dict[DeviceID, DeviceOTKCount]] | None = None,
        device_lists: DeviceLists | None = None,
    ) -> JSON:
        for raw_td in to_device or []:
            try:
                td = ASToDeviceEvent.deserialize(raw_td)
            except SerializerError:
                self.log.exception("Failed to deserialize to-device event %s", raw_td)
            else:
                try:
                    await self.to_device_handler(td)
                except:
                    self.log.exception("Exception in Matrix to-device event handler")
        if device_lists and self.device_list_handler:
            try:
                await self.device_list_handler(device_lists)
            except Exception:
                self.log.exception("Exception in Matrix device list change handler")
        if otk_counts and self.otk_handler:
            try:
                await self.otk_handler(otk_counts)
            except Exception:
                self.log.exception("Exception in Matrix OTK count handler")
        for raw_edu in ephemeral or []:
            try:
                edu = EphemeralEvent.deserialize(raw_edu)
            except SerializerError:
                self.log.exception("Failed to deserialize ephemeral event %s", raw_edu)
            else:
                self.handle_matrix_event(edu, ephemeral=True)
        for raw_event in events:
            try:
                self._fix_prev_content(raw_event)
                event = Event.deserialize(raw_event)
            except SerializerError:
                self.log.exception("Failed to deserialize event %s", raw_event)
            else:
                self.handle_matrix_event(event)
        return {}

    def handle_matrix_event(self, event: Event, ephemeral: bool = False) -> None:
        if ephemeral:
            event.type = event.type.with_class(EventType.Class.EPHEMERAL)
        elif getattr(event, "state_key", None) is not None:
            event.type = event.type.with_class(EventType.Class.STATE)
        else:
            event.type = event.type.with_class(EventType.Class.MESSAGE)

        async def try_handle(handler_func: HandlerFunc):
            try:
                await handler_func(event)
            except Exception:
                self.log.exception("Exception in Matrix event handler")

        for handler in self.event_handlers:
            # TODO add option to handle events synchronously
            background_task.create(try_handle(handler))

    def matrix_event_handler(self, func: HandlerFunc) -> HandlerFunc:
        self.event_handlers.append(func)
        return func



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import asyncio
import logging
import sys

from mautrix import __optional_imports__
from mautrix.appservice import AppService
from mautrix.client import Client, InternalEventType, SyncStore
from mautrix.crypto import CryptoStore, OlmMachine, PgCryptoStore, RejectKeyShare, StateStore
from mautrix.errors import EncryptionError, MForbidden, MNotFound, SessionNotFound
from mautrix.types import (
    JSON,
    DeviceIdentity,
    EncryptedEvent,
    EncryptedMegolmEventContent,
    EventFilter,
    EventType,
    Filter,
    LoginType,
    MessageEvent,
    RequestedKeyInfo,
    RoomEventFilter,
    RoomFilter,
    RoomID,
    RoomKeyWithheldCode,
    Serializable,
    StateEvent,
    StateFilter,
    TrustState,
)
from mautrix.util import background_task
from mautrix.util.async_db import Database
from mautrix.util.logging import TraceLogger

from .. import bridge as br
from .crypto_state_store import PgCryptoStateStore


class EncryptionManager:
    loop: asyncio.AbstractEventLoop
    log: TraceLogger = logging.getLogger("mau.bridge.e2ee")

    client: Client
    crypto: OlmMachine
    crypto_store: CryptoStore | SyncStore
    crypto_db: Database | None
    state_store: StateStore

    min_send_trust: TrustState
    key_sharing_enabled: bool
    appservice_mode: bool
    periodically_delete_expired_keys: bool

    bridge: br.Bridge
    az: AppService
    _id_prefix: str
    _id_suffix: str

    _share_session_events: dict[RoomID, asyncio.Event]
    _key_delete_task: asyncio.Task | None

    def __init__(
        self,
        bridge: br.Bridge,
        homeserver_address: str,
        user_id_prefix: str,
        user_id_suffix: str,
        db_url: str,
    ) -> None:
        self.loop = bridge.loop or asyncio.get_event_loop()
        self.bridge = bridge
        self.az = bridge.az
        self.device_name = bridge.name
        self._id_prefix = user_id_prefix
        self._id_suffix = user_id_suffix
        self._share_session_events = {}
        pickle_key = "mautrix.bridge.e2ee"
        self.crypto_db = Database.create(
            url=db_url,
            upgrade_table=PgCryptoStore.upgrade_table,
            log=logging.getLogger("mau.crypto.db"),
        )
        self.crypto_store = PgCryptoStore("", pickle_key, self.crypto_db)
        self.state_store = PgCryptoStateStore(self.crypto_db, bridge.get_portal)
        default_http_retry_count = bridge.config.get("homeserver.http_retry_count", None)
        self.client = Client(
            base_url=homeserver_address,
            mxid=self.az.bot_mxid,
            loop=self.loop,
            sync_store=self.crypto_store,
            log=self.log.getChild("client"),
            default_retry_count=default_http_retry_count,
            state_store=self.bridge.state_store,
        )
        self.crypto = OlmMachine(self.client, self.crypto_store, self.state_store)
        self.client.add_event_handler(InternalEventType.SYNC_STOPPED, self._exit_on_sync_fail)
        self.crypto.allow_key_share = self.allow_key_share
        verification_levels = bridge.config["bridge.encryption.verification_levels"]
        self.min_send_trust = TrustState.parse(verification_levels["send"])
        self.crypto.share_keys_min_trust = TrustState.parse(verification_levels["share"])
        self.crypto.send_keys_min_trust = TrustState.parse(verification_levels["receive"])
        self.key_sharing_enabled = bridge.config["bridge.encryption.allow_key_sharing"]
        self.appservice_mode = bridge.config["bridge.encryption.appservice"]
        if self.appservice_mode:
            self.az.otk_handler = self.crypto.handle_as_otk_counts
            self.az.device_list_handler = self.crypto.handle_as_device_lists
            self.az.to_device_handler = self.crypto.handle_as_to_device_event

        self.periodically_delete_expired_keys = False
        self._key_delete_task = None
        del_cfg = bridge.config["bridge.encryption.delete_keys"]
        if del_cfg:
            self.crypto.delete_outbound_keys_on_ack = del_cfg["delete_outbound_on_ack"]
            self.crypto.dont_store_outbound_keys = del_cfg["dont_store_outbound"]
            self.crypto.delete_previous_keys_on_receive = del_cfg["delete_prev_on_new_session"]
            self.crypto.ratchet_keys_on_decrypt = del_cfg["ratchet_on_decrypt"]
            self.crypto.delete_fully_used_keys_on_decrypt = del_cfg["delete_fully_used_on_decrypt"]
            self.crypto.delete_keys_on_device_delete = del_cfg["delete_on_device_delete"]
            self.periodically_delete_expired_keys = del_cfg["periodically_delete_expired"]
        self.crypto.disable_device_change_key_rotation = bridge.config[
            "bridge.encryption.rotation.disable_device_change_key_rotation"
        ]

    async def _exit_on_sync_fail(self, data) -> None:
        if data["error"]:
            self.log.critical("Exiting due to crypto sync error")
            sys.exit(32)

    async def allow_key_share(self, device: DeviceIdentity, request: RequestedKeyInfo) -> bool:
        if not self.key_sharing_enabled:
            self.log.debug(
                f"Key sharing not enabled, ignoring key request from "
                f"{device.user_id}/{device.device_id}"
            )
            return False
        elif device.trust == TrustState.BLACKLISTED:
            raise RejectKeyShare(
                f"Rejecting key request from blacklisted device "
                f"{device.user_id}/{device.device_id}",
                code=RoomKeyWithheldCode.BLACKLISTED,
                reason="Your device has been blacklisted by the bridge",
            )
        elif await self.crypto.resolve_trust(device) >= self.crypto.share_keys_min_trust:
            portal = await self.bridge.get_portal(request.room_id)
            if portal is None:
                raise RejectKeyShare(
                    f"Rejecting key request for {request.session_id} from "
                    f"{device.user_id}/{device.device_id}: room is not a portal",
                    code=RoomKeyWithheldCode.UNAVAILABLE,
                    reason="Requested room is not a portal",
                )
            user = await self.bridge.get_user(device.user_id)
            if not await user.is_in_portal(portal):
                raise RejectKeyShare(
                    f"Rejecting key request for {request.session_id} from "
                    f"{device.user_id}/{device.device_id}: user is not in portal",
                    code=RoomKeyWithheldCode.UNAUTHORIZED,
                    reason="You're not in that portal",
                )
            self.log.debug(
                f"Accepting key request for {request.session_id} from "
                f"{device.user_id}/{device.device_id}"
            )
            return True
        else:
            raise RejectKeyShare(
                f"Rejecting key request from unverified device "
                f"{device.user_id}/{device.device_id}",
                code=RoomKeyWithheldCode.UNVERIFIED,
                reason="Your device is not trusted by the bridge",
            )

    def _ignore_user(self, user_id: str) -> bool:
        return (
            user_id.startswith(self._id_prefix)
            and user_id.endswith(self._id_suffix)
            and user_id != self.az.bot_mxid
        )

    async def handle_member_event(self, evt: StateEvent) -> None:
        if self._ignore_user(evt.state_key):
            # We don't want to invalidate group sessions because a ghost left or joined
            return
        await self.crypto.handle_member_event(evt)

    async def _share_session_lock(self, room_id: RoomID) -> bool:
        try:
            event = self._share_session_events[room_id]
        except KeyError:
            self._share_session_events[room_id] = asyncio.Event()
            return True
        else:
            await event.wait()
            return False

    async def encrypt(
        self, room_id: RoomID, event_type: EventType, content: Serializable | JSON
    ) -> tuple[EventType, EncryptedMegolmEventContent]:
        try:
            encrypted = await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        except EncryptionError:
            self.log.debug("Got EncryptionError, sharing group session and trying again")
            if await self._share_session_lock(room_id):
                try:
                    users = await self.az.state_store.get_members_filtered(
                        room_id, self._id_prefix, self._id_suffix, self.az.bot_mxid
                    )
                    await self.crypto.share_group_session(room_id, users)
                finally:
                    self._share_session_events.pop(room_id).set()
            encrypted = await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        return EventType.ROOM_ENCRYPTED, encrypted

    async def decrypt(self, evt: EncryptedEvent, wait_session_timeout: int = 5) -> MessageEvent:
        try:
            decrypted = await self.crypto.decrypt_megolm_event(evt)
        except SessionNotFound as e:
            if not wait_session_timeout:
                raise
            self.log.debug(
                f"Couldn't find session {e.session_id} trying to decrypt {evt.event_id},"
                f" waiting {wait_session_timeout} seconds..."
            )
            got_keys = await self.crypto.wait_for_session(
                evt.room_id, e.session_id, timeout=wait_session_timeout
            )
            if got_keys:
                self.log.debug(
                    f"Got session {e.session_id} after waiting, "
                    f"trying to decrypt {evt.event_id} again"
                )
                decrypted = await self.crypto.decrypt_megolm_event(evt)
            else:
                raise
        self.log.trace("Decrypted event %s: %s", evt.event_id, decrypted)
        return decrypted

    async def start(self) -> None:
        flows = await self.client.get_login_flows()
        if not flows.supports_type(LoginType.APPSERVICE):
            self.log.critical(
                "Encryption enabled in config, but homeserver does not support appservice login"
            )
            sys.exit(30)
        self.log.debug("Logging in with bridge bot user")
        if self.crypto_db:
            try:
                await self.crypto_db.start()
            except Exception as e:
                self.bridge._log_db_error(e)
        await self.crypto_store.open()
        device_id = await self.crypto_store.get_device_id()
        if device_id:
            self.log.debug(f"Found device ID in database: {device_id}")
        # We set the API token to the AS token here to authenticate the appservice login
        # It'll get overridden after the login
        self.client.api.token = self.az.as_token
        await self.client.login(
            login_type=LoginType.APPSERVICE,
            device_name=self.device_name,
            device_id=device_id,
            store_access_token=True,
            update_hs_url=False,
        )
        await self.crypto.load()
        if not device_id:
            await self.crypto_store.put_device_id(self.client.device_id)
            self.log.debug(f"Logged in with new device ID {self.client.device_id}")
        elif self.crypto.account.shared:
            await self._verify_keys_are_on_server()
        if self.appservice_mode:
            self.log.info("End-to-bridge encryption support is enabled (appservice mode)")
        else:
            _ = self.client.start(self._filter)
            self.log.info("End-to-bridge encryption support is enabled (sync mode)")
        if self.periodically_delete_expired_keys:
            self._key_delete_task = background_task.create(self._periodically_delete_keys())
        background_task.create(self._resync_encryption_info())

    async def _resync_encryption_info(self) -> None:
        rows = await self.crypto_db.fetch(
            """SELECT room_id FROM mx_room_state WHERE encryption='{"resync":true}'"""
        )
        room_ids = [row["room_id"] for row in rows]
        if not room_ids:
            return
        self.log.debug(f"Resyncing encryption state event in rooms: {room_ids}")
        for room_id in room_ids:
            try:
                evt = await self.client.get_state_event(room_id, EventType.ROOM_ENCRYPTION)
            except (MNotFound, MForbidden) as e:
                self.log.debug(f"Failed to get encryption state in {room_id}: {e}")
                q = """
                    UPDATE mx_room_state SET encryption=NULL
                    WHERE room_id=$1 AND encryption='{"resync":true}'
                """
                await self.crypto_db.execute(q, room_id)
            else:
                self.log.debug(f"Resynced encryption state in {room_id}: {evt}")
                q = """
                    UPDATE crypto_megolm_inbound_session SET max_age=$1, max_messages=$2
                    WHERE room_id=$3 AND max_age IS NULL and max_messages IS NULL
                """
                await self.crypto_db.execute(
                    q, evt.rotation_period_ms, evt.rotation_period_msgs, room_id
                )

    async def _verify_keys_are_on_server(self) -> None:
        self.log.debug("Making sure keys are still on server")
        try:
            resp = await self.client.query_keys([self.client.mxid])
        except Exception:
            self.log.critical(
                "Failed to query own keys to make sure device still exists", exc_info=True
            )
            sys.exit(33)
        try:
            own_keys = resp.device_keys[self.client.mxid][self.client.device_id]
            if len(own_keys.keys) > 0:
                return
        except KeyError:
            pass
        self.log.critical("Existing device doesn't have keys on server, resetting crypto")
        await self.crypto.crypto_store.delete()
        await self.client.logout_all()
        sys.exit(34)

    async def stop(self) -> None:
        if self._key_delete_task:
            self._key_delete_task.cancel()
            self._key_delete_task = None
        self.client.stop()
        await self.crypto_store.close()
        if self.crypto_db:
            await self.crypto_db.stop()

    @property
    def _filter(self) -> Filter:
        all_events = EventType.find("*")
        return Filter(
            account_data=EventFilter(types=[all_events]),
            presence=EventFilter(not_types=[all_events]),
            room=RoomFilter(
                include_leave=False,
                state=StateFilter(not_types=[all_events]),
                timeline=RoomEventFilter(not_types=[all_events]),
                account_data=RoomEventFilter(not_types=[all_events]),
                ephemeral=RoomEventFilter(not_types=[all_events]),
            ),
        )

    async def _periodically_delete_keys(self) -> None:
        while True:
            deleted = await self.crypto_store.redact_expired_group_sessions()
            if deleted:
                self.log.info(f"Deleted expired megolm sessions: {deleted}")
            else:
                self.log.debug("No expired megolm sessions found")
            await asyncio.sleep(24 * 60 * 60)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging

from mautrix.util.async_db import Connection, Scheme, UpgradeTable

upgrade_table = UpgradeTable(
    version_table_name="mx_version",
    database_name="matrix state cache",
    log=logging.getLogger("mau.client.db.upgrade"),
)


@upgrade_table.register(description="Latest revision", upgrades_to=3)
async def upgrade_blank_to_v3(conn: Connection, scheme: Scheme) -> None:
    await conn.execute(
        """CREATE TABLE mx_room_state (
            room_id              TEXT PRIMARY KEY,
            is_encrypted         BOOLEAN,
            has_full_member_list BOOLEAN,
            encryption           TEXT,
            power_levels         TEXT
        )"""
    )
    membership_check = ""
    if scheme != Scheme.SQLITE:
        await conn.execute(
            "CREATE TYPE membership AS ENUM ('join', 'leave', 'invite', 'ban', 'knock')"
        )
    else:
        membership_check = "CHECK (membership IN ('join', 'leave', 'invite', 'ban', 'knock'))"
    await conn.execute(
        f"""CREATE TABLE mx_user_profile (
            room_id     TEXT,
            user_id     TEXT,
            membership  membership NOT NULL {membership_check},
            displayname TEXT,
            avatar_url  TEXT,
            PRIMARY KEY (room_id, user_id)
        )"""
    )


@upgrade_table.register(description="Stop using size-limited string fields")
async def upgrade_v2(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.SQLITE:
        # SQLite doesn't care about types
        return
    await conn.execute("ALTER TABLE mx_room_state ALTER COLUMN room_id TYPE TEXT")
    await conn.execute("ALTER TABLE mx_user_profile ALTER COLUMN room_id TYPE TEXT")
    await conn.execute("ALTER TABLE mx_user_profile ALTER COLUMN user_id TYPE TEXT")
    await conn.execute("ALTER TABLE mx_user_profile ALTER COLUMN displayname TYPE TEXT")
    await conn.execute("ALTER TABLE mx_user_profile ALTER COLUMN avatar_url TYPE TEXT")


@upgrade_table.register(description="Mark rooms that need crypto state event resynced")
async def upgrade_v3(conn: Connection) -> None:
    if await conn.table_exists("portal"):
        await conn.execute(
            """
            INSERT INTO mx_room_state (room_id, encryption)
            SELECT portal.mxid, '{"resync":true}' FROM portal
                WHERE portal.encrypted=true AND portal.mxid IS NOT NULL
            ON CONFLICT (room_id) DO UPDATE
                SET encryption=excluded.encryption
                WHERE mx_room_state.encryption IS NULL
            """
        )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations


class DatabaseException(RuntimeError):
    pass

    @property
    def explanation(self) -> str | None:
        return None


class UnsupportedDatabaseVersion(DatabaseException):
    def __init__(self, name: str, version: int, latest: int) -> None:
        super().__init__(
            f"Unsupported {name} schema version v{version} (latest known is v{latest})"
        )

    @property
    def explanation(self) -> str:
        return "Downgrading is not supported"


class ForeignTablesFound(DatabaseException):
    def __init__(self, explanation: str) -> None:
        super().__init__(f"The database contains foreign tables ({explanation})")

    @property
    def explanation(self) -> str:
        return "You can use --ignore-foreign-tables to ignore this error"


class DatabaseNotOwned(DatabaseException):
    def __init__(self, owner: str) -> None:
        super().__init__(f"The database is owned by {owner}")

    @property
    def explanation(self) -> str:
        return "Sharing the same database with different programs is not supported"



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Generic, Type, TypeVar
import copy

from ruamel.yaml.comments import CommentedMap

T = TypeVar("T")


class RecursiveDict(Generic[T]):
    def __init__(self, data: T | None = None, dict_factory: Type[T] | None = None) -> None:
        self._dict_factory = dict_factory or dict
        self._data: CommentedMap = data or self._dict_factory()

    def clone(self) -> RecursiveDict:
        return RecursiveDict(data=copy.deepcopy(self._data), dict_factory=self._dict_factory)

    @staticmethod
    def parse_key(key: str) -> tuple[str, str | None]:
        if "." not in key:
            return key, None
        key, next_key = key.split(".", 1)
        if len(key) > 0 and key[0] == "[":
            end_index = next_key.index("]")
            key = key[1:] + "." + next_key[:end_index]
            next_key = next_key[end_index + 2 :] if len(next_key) > end_index + 1 else None
        return key, next_key

    def _recursive_get(self, data: T, key: str, default_value: Any) -> Any:
        key, next_key = self.parse_key(key)
        if next_key is not None:
            next_data = data.get(key, self._dict_factory())
            return self._recursive_get(next_data, next_key, default_value)
        try:
            return data[key]
        except (AttributeError, KeyError):
            return default_value

    def get(self, key: str, default_value: Any, allow_recursion: bool = True) -> Any:
        if allow_recursion and "." in key:
            return self._recursive_get(self._data, key, default_value)
        return self._data.get(key, default_value)

    def __getitem__(self, key: str) -> Any:
        return self.get(key, None)

    def __contains__(self, key: str) -> bool:
        return self.get(key, None) is not None

    def _recursive_set(self, data: T, key: str, value: Any) -> None:
        key, next_key = self.parse_key(key)
        if next_key is not None:
            if key not in data:
                data[key] = self._dict_factory()
            next_data = data.get(key, self._dict_factory())
            return self._recursive_set(next_data, next_key, value)
        data[key] = value

    def set(self, key: str, value: Any, allow_recursion: bool = True) -> None:
        if allow_recursion and "." in key:
            self._recursive_set(self._data, key, value)
            return
        self._data[key] = value

    def __setitem__(self, key: str, value: Any) -> None:
        self.set(key, value)

    def _recursive_del(self, data: T, key: str) -> None:
        key, next_key = self.parse_key(key)
        if next_key is not None:
            if key not in data:
                return
            next_data = data[key]
            return self._recursive_del(next_data, next_key)
        try:
            del data[key]
            del data.ca.items[key]
        except KeyError:
            pass

    def delete(self, key: str, allow_recursion: bool = True) -> None:
        if allow_recursion and "." in key:
            self._recursive_del(self._data, key)
            return
        try:
            del self._data[key]
            del self._data.ca.items[key]
        except KeyError:
            pass

    def __delitem__(self, key: str) -> None:
        self.delete(key)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, AsyncContextManager
from contextlib import asynccontextmanager
import asyncio
import logging
import os
import re
import sqlite3

from yarl import URL
import aiosqlite

from .connection import LoggingConnection
from .database import Database
from .scheme import Scheme
from .upgrade import UpgradeTable

POSITIONAL_PARAM_PATTERN = re.compile(r"\$(\d+)")


class TxnConnection(aiosqlite.Connection):
    def __init__(self, path: str, **kwargs) -> None:
        def connector() -> sqlite3.Connection:
            return sqlite3.connect(
                path, detect_types=sqlite3.PARSE_DECLTYPES, isolation_level=None, **kwargs
            )

        super().__init__(connector, iter_chunk_size=64)

    @asynccontextmanager
    async def transaction(self) -> None:
        await self.execute("BEGIN TRANSACTION")
        try:
            yield
        except Exception:
            await self.rollback()
            raise
        else:
            await self.commit()

    def __execute(self, query: str, *args: Any):
        query = POSITIONAL_PARAM_PATTERN.sub(r"?\1", query)
        return super().execute(query, args)

    async def execute(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> aiosqlite.Cursor:
        return await self.__execute(query, *args)

    async def executemany(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> aiosqlite.Cursor:
        query = POSITIONAL_PARAM_PATTERN.sub(r"?\1", query)
        return await super().executemany(query, *args)

    async def fetch(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> list[sqlite3.Row]:
        async with self.__execute(query, *args) as cursor:
            return list(await cursor.fetchall())

    async def fetchrow(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> sqlite3.Row | None:
        async with self.__execute(query, *args) as cursor:
            return await cursor.fetchone()

    async def fetchval(
        self, query: str, *args: Any, column: int = 0, timeout: float | None = None
    ) -> Any:
        row = await self.fetchrow(query, *args)
        if row is None:
            return None
        return row[column]


class SQLiteDatabase(Database):
    scheme = Scheme.SQLITE
    _parent: SQLiteDatabase | None
    _pool: asyncio.Queue[TxnConnection]
    _stopped: bool
    _conns: int
    _init_commands: list[str]

    def __init__(
        self,
        url: URL,
        upgrade_table: UpgradeTable,
        db_args: dict[str, Any] | None = None,
        log: logging.Logger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> None:
        super().__init__(
            url,
            db_args=db_args,
            upgrade_table=upgrade_table,
            log=log,
            owner_name=owner_name,
            ignore_foreign_tables=ignore_foreign_tables,
        )
        self._parent = None
        self._path = url.path
        self._pool = asyncio.Queue(self._db_args.pop("min_size", 1))
        self._db_args.pop("max_size", None)
        self._stopped = False
        self._conns = 0
        self._init_commands = self._add_missing_pragmas(self._db_args.pop("init_commands", []))

    @staticmethod
    def _add_missing_pragmas(init_commands: list[str]) -> list[str]:
        has_foreign_keys = False
        has_journal_mode = False
        has_busy_timeout = False
        for cmd in init_commands:
            if "PRAGMA" not in cmd:
                continue
            if "foreign_keys" in cmd:
                has_foreign_keys = True
            elif "journal_mode" in cmd:
                has_journal_mode = True
            elif "busy_timeout" in cmd:
                has_busy_timeout = True
        if not has_foreign_keys:
            init_commands.append("PRAGMA foreign_keys = ON")
        if not has_journal_mode:
            init_commands.append("PRAGMA journal_mode = WAL")
        if not has_busy_timeout:
            init_commands.append("PRAGMA busy_timeout = 5000")
        return init_commands

    def override_pool(self, db: Database) -> None:
        assert isinstance(db, SQLiteDatabase)
        self._parent = db

    async def start(self) -> None:
        if self._parent:
            await super().start()
            return
        if self._conns:
            raise RuntimeError("database pool has already been started")
        elif self._stopped:
            raise RuntimeError("database pool can't be restarted")
        self.log.debug(f"Connecting to {self.url}")
        self.log.debug(f"Database connection init commands: {self._init_commands}")
        if os.path.exists(self._path):
            if not os.access(self._path, os.W_OK):
                self.log.warning("Database file doesn't seem writable")
        elif not os.access(os.path.dirname(os.path.abspath(self._path)), os.W_OK):
            self.log.warning("Database file doesn't exist and directory doesn't seem writable")
        for _ in range(self._pool.maxsize):
            conn = await TxnConnection(self._path, **self._db_args)
            if self._init_commands:
                cur = await conn.cursor()
                for command in self._init_commands:
                    self.log.trace("Executing init command: %s", command)
                    await cur.execute(command)
                await conn.commit()
            conn.row_factory = sqlite3.Row
            self._pool.put_nowait(conn)
            self._conns += 1
        await super().start()

    async def stop(self) -> None:
        if self._parent:
            return
        self._stopped = True
        while self._conns > 0:
            conn = await self._pool.get()
            self._conns -= 1
            await conn.close()

    def acquire(self) -> AsyncContextManager[LoggingConnection]:
        if self._parent:
            return self._parent.acquire()
        return self._acquire()

    @asynccontextmanager
    async def _acquire(self) -> LoggingConnection:
        if self._stopped:
            raise RuntimeError("database pool has been stopped")
        conn = await self._pool.get()
        try:
            yield LoggingConnection(self.scheme, conn, self.log)
        finally:
            self._pool.put_nowait(conn)


Database.schemes["sqlite"] = SQLiteDatabase
Database.schemes["sqlite3"] = SQLiteDatabase



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Dict, List, NamedTuple, Optional
from enum import IntEnum

from attr import dataclass

from .event import EncryptionAlgorithm, EncryptionKeyAlgorithm, KeyID, ToDeviceEvent
from .primitive import DeviceID, IdentityKey, Signature, SigningKey, UserID
from .util import ExtensibleEnum, SerializableAttrs, field


@dataclass
class UnsignedDeviceInfo(SerializableAttrs):
    device_display_name: Optional[str] = None


@dataclass
class DeviceKeys(SerializableAttrs):
    user_id: UserID
    device_id: DeviceID
    algorithms: List[EncryptionAlgorithm]
    keys: Dict[KeyID, str]
    signatures: Dict[UserID, Dict[KeyID, Signature]]
    unsigned: UnsignedDeviceInfo = None

    def __attrs_post_init__(self) -> None:
        if self.unsigned is None:
            self.unsigned = UnsignedDeviceInfo()

    @property
    def ed25519(self) -> Optional[SigningKey]:
        try:
            return SigningKey(self.keys[KeyID(EncryptionKeyAlgorithm.ED25519, self.device_id)])
        except KeyError:
            return None

    @property
    def curve25519(self) -> Optional[IdentityKey]:
        try:
            return IdentityKey(self.keys[KeyID(EncryptionKeyAlgorithm.CURVE25519, self.device_id)])
        except KeyError:
            return None


class CrossSigningUsage(ExtensibleEnum):
    MASTER = "master"
    SELF = "self_signing"
    USER = "user_signing"


@dataclass
class CrossSigningKeys(SerializableAttrs):
    user_id: UserID
    usage: List[CrossSigningUsage]
    keys: Dict[KeyID, SigningKey]
    signatures: Dict[UserID, Dict[KeyID, Signature]] = field(factory=lambda: {})

    @property
    def first_key(self) -> Optional[SigningKey]:
        try:
            return next(iter(self.keys.values()))
        except StopIteration:
            return None

    @property
    def first_ed25519_key(self) -> Optional[SigningKey]:
        return self.first_key_with_algorithm(EncryptionKeyAlgorithm.ED25519)

    def first_key_with_algorithm(self, alg: EncryptionKeyAlgorithm) -> Optional[SigningKey]:
        if not self.keys:
            return None
        try:
            return next(key for key_id, key in self.keys.items() if key_id.algorithm == alg)
        except StopIteration:
            return None


@dataclass
class QueryKeysResponse(SerializableAttrs):
    device_keys: Dict[UserID, Dict[DeviceID, DeviceKeys]] = field(factory=lambda: {})
    master_keys: Dict[UserID, CrossSigningKeys] = field(factory=lambda: {})
    self_signing_keys: Dict[UserID, CrossSigningKeys] = field(factory=lambda: {})
    user_signing_keys: Dict[UserID, CrossSigningKeys] = field(factory=lambda: {})
    failures: Dict[str, Any] = field(factory=lambda: {})


@dataclass
class ClaimKeysResponse(SerializableAttrs):
    one_time_keys: Dict[UserID, Dict[DeviceID, Dict[KeyID, Any]]]
    failures: Dict[str, Any] = field(factory=lambda: {})


class TrustState(IntEnum):
    BLACKLISTED = -100
    UNVERIFIED = 0
    UNKNOWN_DEVICE = 10
    FORWARDED = 20
    CROSS_SIGNED_UNTRUSTED = 50
    CROSS_SIGNED_TOFU = 100
    CROSS_SIGNED_TRUSTED = 200
    VERIFIED = 300

    def __str__(self) -> str:
        return _trust_state_to_name[self]

    @classmethod
    def parse(cls, val: str) -> "TrustState":
        try:
            return _name_to_trust_state[val]
        except KeyError as e:
            raise ValueError(f"Invalid trust state {val!r}") from e


_trust_state_to_name: Dict[TrustState, str] = {
    val: val.name.lower().replace("_", "-") for val in TrustState
}
_name_to_trust_state: Dict[str, TrustState] = {
    value: key for key, value in _trust_state_to_name.items()
}


@dataclass
class DeviceIdentity:
    user_id: UserID
    device_id: DeviceID
    identity_key: IdentityKey
    signing_key: SigningKey

    trust: TrustState
    deleted: bool
    name: str


@dataclass
class OlmEventKeys(SerializableAttrs):
    ed25519: SigningKey


@dataclass
class DecryptedOlmEvent(ToDeviceEvent, SerializableAttrs):
    keys: OlmEventKeys
    recipient: UserID
    recipient_keys: OlmEventKeys
    sender_device: Optional[DeviceID] = None
    sender_key: IdentityKey = field(hidden=True, default=None)


class TOFUSigningKey(NamedTuple):
    """
    A tuple representing a single cross-signing key. The first value is the current key, and the
    second value is the first seen key. If the values don't match, it means the key is not valid
    for trust-on-first-use.
    """

    key: SigningKey
    first: SigningKey


class CrossSigner(NamedTuple):
    """
    A tuple containing a user ID and a signing key they own.

    The key can either be a device-owned signing key, or one of the user's cross-signing keys.
    """

    user_id: UserID
    key: SigningKey



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import AsyncContextManager, AsyncIterator, Callable
from contextlib import asynccontextmanager
import os
import random
import string
import time

import asyncpg
import pytest

from mautrix.client.state_store import SyncStore
from mautrix.crypto import InboundGroupSession, OlmAccount, OutboundGroupSession
from mautrix.types import DeviceID, EventID, RoomID, SessionID, SyncToken
from mautrix.util.async_db import Database

from .. import CryptoStore, MemoryCryptoStore, PgCryptoStore


@asynccontextmanager
async def async_postgres_store() -> AsyncIterator[PgCryptoStore]:
    try:
        pg_url = os.environ["MEOW_TEST_PG_URL"]
    except KeyError:
        pytest.skip("Skipped Postgres tests (MEOW_TEST_PG_URL not specified)")
        return
    conn: asyncpg.Connection = await asyncpg.connect(pg_url)
    schema_name = "".join(random.choices(string.ascii_lowercase, k=8))
    schema_name = f"test_schema_{schema_name}_{int(time.time())}"
    await conn.execute(f"CREATE SCHEMA {schema_name}")
    db = Database.create(
        pg_url,
        upgrade_table=PgCryptoStore.upgrade_table,
        db_args={"min_size": 1, "max_size": 3, "server_settings": {"search_path": schema_name}},
    )
    store = PgCryptoStore("", "test", db)
    await db.start()
    yield store
    await db.stop()
    await conn.execute(f"DROP SCHEMA {schema_name} CASCADE")
    await conn.close()


@asynccontextmanager
async def async_sqlite_store() -> AsyncIterator[PgCryptoStore]:
    db = Database.create(
        "sqlite::memory:", upgrade_table=PgCryptoStore.upgrade_table, db_args={"min_size": 1}
    )
    store = PgCryptoStore("", "test", db)
    await db.start()
    yield store
    await db.stop()


@asynccontextmanager
async def memory_store() -> AsyncIterator[MemoryCryptoStore]:
    yield MemoryCryptoStore("", "test")


@pytest.fixture(params=[async_postgres_store, async_sqlite_store, memory_store])
async def crypto_store(request) -> AsyncIterator[CryptoStore]:
    param: Callable[[], AsyncContextManager[CryptoStore]] = request.param
    async with param() as state_store:
        yield state_store


async def test_basic(crypto_store: CryptoStore) -> None:
    acc = OlmAccount()
    keys = acc.identity_keys
    await crypto_store.put_account(acc)
    await crypto_store.put_device_id(DeviceID("TEST"))
    if isinstance(crypto_store, SyncStore):
        await crypto_store.put_next_batch(SyncToken("TEST"))

    assert await crypto_store.get_device_id() == "TEST"
    assert (await crypto_store.get_account()).identity_keys == keys
    if isinstance(crypto_store, SyncStore):
        assert await crypto_store.get_next_batch() == "TEST"


def _make_group_sess(
    acc: OlmAccount, room_id: RoomID
) -> tuple[InboundGroupSession, OutboundGroupSession]:
    outbound = OutboundGroupSession(room_id)
    inbound = InboundGroupSession(
        session_key=outbound.session_key,
        signing_key=acc.signing_key,
        sender_key=acc.identity_key,
        room_id=room_id,
    )
    return inbound, outbound


async def test_validate_message_index(crypto_store: CryptoStore) -> None:
    acc = OlmAccount()

    inbound, outbound = _make_group_sess(acc, RoomID("!foo:bar.com"))
    outbound.shared = True
    orig_plaintext = "hello world"
    ciphertext = outbound.encrypt(orig_plaintext)
    ts = int(time.time() * 1000)
    plaintext, index = inbound.decrypt(ciphertext)
    assert plaintext == orig_plaintext

    assert await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$foo"), index, ts
    ), "Initial validation returns True"
    assert await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$foo"), index, ts
    ), "Validating the same details again returns True"
    assert not await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$bar"), index, ts
    ), "Different event ID causes validation to fail"
    assert not await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$foo"), index, ts + 1
    ), "Different timestamp causes validation to fail"
    assert not await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$foo"), index, ts + 1
    ), "Validating incorrect details twice fails"
    assert await crypto_store.validate_message_index(
        acc.identity_key, SessionID(inbound.id), EventID("$foo"), index, ts
    ), "Validating the same details after fails still returns True"


# TODO tests for device identity storage, group session storage
#      and cross-signing key/signature storage



__all__ = [
    # Directory modules
    "async_db",
    "config",
    "db",
    "formatter",
    "logging",
    # File modules
    "async_body",
    "async_getter_lock",
    "background_task",
    "bridge_state",
    "color_log",
    "ffmpeg",
    "file_store",
    "format_duration",
    "magic",
    "manhole",
    "markdown",
    "message_send_checkpoint",
    "opt_prometheus",
    "program",
    "signed_token",
    "simple_lock",
    "simple_template",
    "utf16_surrogate",
    "variation_selector",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import List, NamedTuple, Optional, Union

from mautrix.appservice import IntentAPI
from mautrix.errors import MatrixRequestError
from mautrix.types import EventID, EventType, RoomID, UserID

from ... import bridge as br
from .handler import SECTION_ADMIN, CommandEvent, command_handler


class ManagementRoom(NamedTuple):
    room_id: RoomID
    user_id: UserID


class RoomSearchResults(NamedTuple):
    management_rooms: List[ManagementRoom]
    unidentified_rooms: List[RoomID]
    tombstoned_rooms: List[RoomID]
    portals: List[br.BasePortal]
    empty_portals: List[br.BasePortal]


async def _find_rooms(bridge: br.Bridge, intent: Optional[IntentAPI] = None) -> RoomSearchResults:
    results = RoomSearchResults([], [], [], [], [])
    intent = intent or bridge.az.intent
    rooms = await intent.get_joined_rooms()
    for room_id in rooms:
        portal = await bridge.get_portal(room_id)
        if not portal:
            try:
                tombstone = await intent.get_state_event(room_id, EventType.ROOM_TOMBSTONE)
                if tombstone and tombstone.replacement_room:
                    results.tombstoned_rooms.append(room_id)
                    continue
            except MatrixRequestError:
                pass
            try:
                members = await intent.get_room_members(room_id)
            except MatrixRequestError:
                members = []
            if len(members) == 2:
                other_member = members[0] if members[0] != intent.mxid else members[1]
                if bridge.is_bridge_ghost(other_member):
                    results.unidentified_rooms.append(room_id)
                else:
                    results.management_rooms.append(ManagementRoom(room_id, other_member))
            else:
                results.unidentified_rooms.append(room_id)
        else:
            members = await portal.get_authenticated_matrix_users()
            if len(members) == 0:
                results.empty_portals.append(portal)
            else:
                results.portals.append(portal)

    return results


@command_handler(
    needs_admin=True,
    needs_auth=False,
    management_only=True,
    name="clean-rooms",
    help_section=SECTION_ADMIN,
    help_text="Clean up unused portal/management rooms.",
)
async def clean_rooms(evt: CommandEvent) -> EventID:
    results = await _find_rooms(evt.bridge)

    reply = ["#### Management rooms (M)"]
    reply += [
        f"{n + 1}. [M{n + 1}](https://matrix.to/#/{room}) (with {other_member}"
        for n, (room, other_member) in enumerate(results.management_rooms)
    ] or ["No management rooms found."]
    reply.append("#### Active portal rooms (A)")
    reply += [
        f"{n + 1}. [A{n + 1}](https://matrix.to/#/{portal.mxid}) "
        f'(to remote chat "{portal.name}")'
        for n, portal in enumerate(results.portals)
    ] or ["No active portal rooms found."]
    reply.append("#### Unidentified rooms (U)")
    reply += [
        f"{n + 1}. [U{n + 1}](https://matrix.to/#/{room})"
        for n, room in enumerate(results.unidentified_rooms)
    ] or ["No unidentified rooms found."]
    reply.append("#### Tombstoned rooms (T)")
    reply += [
        f"{n + 1}. [T{n + 1}](https://matrix.to/#/{room})"
        for n, room in enumerate(results.tombstoned_rooms)
    ] or ["No tombstoned rooms found."]
    reply.append("#### Inactive portal rooms (I)")
    reply += [
        f'{n}. [I{n}](https://matrix.to/#/{portal.mxid}) (to remote chat "{portal.name}")'
        for n, portal in enumerate(results.empty_portals)
    ] or ["No inactive portal rooms found."]

    reply += [
        "#### Usage",
        (
            "To clean the recommended set of rooms (unidentified & inactive portals), "
            "type `$cmdprefix+sp clean-recommended`"
        ),
        "",
        (
            "To clean other groups of rooms, type `$cmdprefix+sp clean-groups <letters>` "
            "where `letters` are the first letters of the group names (M, A, U, I, T)"
        ),
        "",
        (
            "To clean specific rooms, type `$cmdprefix+sp clean-range <range>` "
            "where `range` is the range (e.g. `5-21`) prefixed with the first letter of"
            "the group name. (e.g. `I2-6`)"
        ),
        "",
        (
            "Please note that you will have to re-run `$cmdprefix+sp clean-rooms` "
            "between each use of the commands above."
        ),
    ]

    evt.sender.command_status = {
        "next": lambda clean_evt: set_rooms_to_clean(clean_evt, results),
        "action": "Room cleaning",
    }

    return await evt.reply("\n".join(reply))


async def set_rooms_to_clean(evt, results: RoomSearchResults) -> None:
    command = evt.args[0]
    rooms_to_clean: List[Union[br.BasePortal, RoomID]] = []
    if command == "clean-recommended":
        rooms_to_clean += results.empty_portals
        rooms_to_clean += results.unidentified_rooms
    elif command == "clean-groups":
        if len(evt.args) < 2:
            return await evt.reply("**Usage:** $cmdprefix+sp clean-groups [M][A][U][I]")
        groups_to_clean = evt.args[1].upper()
        if "M" in groups_to_clean:
            rooms_to_clean += [room_id for (room_id, user_id) in results.management_rooms]
        if "A" in groups_to_clean:
            rooms_to_clean += results.portals
        if "U" in groups_to_clean:
            rooms_to_clean += results.unidentified_rooms
        if "I" in groups_to_clean:
            rooms_to_clean += results.empty_portals
        if "T" in groups_to_clean:
            rooms_to_clean += results.tombstoned_rooms
    elif command == "clean-range":
        try:
            clean_range = evt.args[1]
            group, clean_range = clean_range[0], clean_range[1:]
            start, end = clean_range.split("-")
            start, end = int(start), int(end)
            if group == "M":
                group = [room_id for (room_id, user_id) in results.management_rooms]
            elif group == "A":
                group = results.portals
            elif group == "U":
                group = results.unidentified_rooms
            elif group == "I":
                group = results.empty_portals
            elif group == "T":
                group = results.tombstoned_rooms
            else:
                raise ValueError("Unknown group")
            rooms_to_clean = group[start - 1 : end]
        except (KeyError, ValueError):
            return await evt.reply("**Usage:** $cmdprefix+sp clean-range <_M|A|U|I_><range>")
    else:
        return await evt.reply(
            f"Unknown room cleaning action `{command}`. "
            "Use `$cmdprefix+sp cancel` to cancel room cleaning."
        )

    evt.sender.command_status = {
        "next": lambda confirm: execute_room_cleanup(confirm, rooms_to_clean),
        "action": "Room cleaning",
    }
    await evt.reply(
        f"To confirm cleaning up {len(rooms_to_clean)} rooms, type `$cmdprefix+sp confirm-clean`."
    )


async def execute_room_cleanup(evt, rooms_to_clean: List[Union[br.BasePortal, RoomID]]) -> None:
    if len(evt.args) > 0 and evt.args[0] == "confirm-clean":
        await evt.reply(f"Cleaning {len(rooms_to_clean)} rooms. This might take a while.")
        cleaned = 0
        for room in rooms_to_clean:
            if isinstance(room, br.BasePortal):
                await room.cleanup_and_delete()
                cleaned += 1
            else:
                await br.BasePortal.cleanup_room(evt.az.intent, room, "Room deleted")
                cleaned += 1
        evt.sender.command_status = None
        await evt.reply(f"{cleaned} rooms cleaned up successfully.")
    else:
        await evt.reply("Room cleaning cancelled.")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable, Callable, TypeVar
from contextlib import asynccontextmanager
from logging import WARNING
import functools
import time

from mautrix import __optional_imports__
from mautrix.util.logging import SILLY, TraceLogger

from .scheme import Scheme

if __optional_imports__:
    from sqlite3 import Row

    from aiosqlite import Cursor
    from asyncpg import Record
    import asyncpg

    from . import aiosqlite

Decorated = TypeVar("Decorated", bound=Callable[..., Any])
LOG_MESSAGE = "%s(%r) took %.3f seconds"


def log_duration(func: Decorated) -> Decorated:
    func_name = func.__name__

    @functools.wraps(func)
    async def wrapper(self: LoggingConnection, arg: str, *args: Any, **kwargs: str) -> Any:
        start = time.monotonic()
        ret = await func(self, arg, *args, **kwargs)
        duration = time.monotonic() - start
        self.log.log(WARNING if duration > 1 else SILLY, LOG_MESSAGE, func_name, arg, duration)
        return ret

    return wrapper


async def handle_exception_noop(_: Exception) -> None:
    pass


class LoggingConnection:
    def __init__(
        self,
        scheme: Scheme,
        wrapped: aiosqlite.TxnConnection | asyncpg.Connection,
        log: TraceLogger,
        handle_exception: Callable[[Exception], Awaitable[None]] = handle_exception_noop,
    ) -> None:
        self.scheme = scheme
        self.wrapped = wrapped
        self.log = log
        self._handle_exception = handle_exception
        self._inited = True

    def __setattr__(self, key: str, value: Any) -> None:
        if getattr(self, "_inited", False):
            raise RuntimeError("LoggingConnection fields are frozen")
        super().__setattr__(key, value)

    @asynccontextmanager
    async def transaction(self) -> None:
        try:
            async with self.wrapped.transaction():
                yield
        except Exception as e:
            await self._handle_exception(e)
            raise

    @log_duration
    async def execute(self, query: str, *args: Any, timeout: float | None = None) -> str | Cursor:
        try:
            return await self.wrapped.execute(query, *args, timeout=timeout)
        except Exception as e:
            await self._handle_exception(e)
            raise

    @log_duration
    async def executemany(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> str | Cursor:
        try:
            return await self.wrapped.executemany(query, *args, timeout=timeout)
        except Exception as e:
            await self._handle_exception(e)
            raise

    @log_duration
    async def fetch(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> list[Row | Record]:
        try:
            return await self.wrapped.fetch(query, *args, timeout=timeout)
        except Exception as e:
            await self._handle_exception(e)
            raise

    @log_duration
    async def fetchval(
        self, query: str, *args: Any, column: int = 0, timeout: float | None = None
    ) -> Any:
        try:
            return await self.wrapped.fetchval(query, *args, column=column, timeout=timeout)
        except Exception as e:
            await self._handle_exception(e)
            raise

    @log_duration
    async def fetchrow(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> Row | Record | None:
        try:
            return await self.wrapped.fetchrow(query, *args, timeout=timeout)
        except Exception as e:
            await self._handle_exception(e)
            raise

    async def table_exists(self, name: str) -> bool:
        if self.scheme == Scheme.SQLITE:
            return await self.fetchval(
                "SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE type='table' AND name=?1)", name
            )
        elif self.scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
            return await self.fetchval(
                "SELECT EXISTS(SELECT 1 FROM information_schema.tables WHERE table_name=$1)", name
            )
        else:
            raise RuntimeError(f"Unknown scheme {self.scheme}")

    @log_duration
    async def copy_records_to_table(
        self,
        table_name: str,
        *,
        records: list[tuple[Any, ...]],
        columns: tuple[str, ...] | list[str],
        schema_name: str | None = None,
        timeout: float | None = None,
    ) -> None:
        if self.scheme != Scheme.POSTGRES:
            raise RuntimeError("copy_records_to_table is only supported on Postgres")
        try:
            return await self.wrapped.copy_records_to_table(
                table_name,
                records=records,
                columns=columns,
                schema_name=schema_name,
                timeout=timeout,
            )
        except Exception as e:
            await self._handle_exception(e)
            raise



# Copyright 2018 Zil0 (under the Apache 2.0 license)
# Copyright © 2019 Damir Jelić <poljar@termina.org.uk> (under the Apache 2.0 license)
# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

from __future__ import annotations

from typing import Generator, Iterable
import binascii
import struct

import unpaddedbase64

from mautrix.errors import DecryptionError
from mautrix.types import EncryptedFile, JSONWebKey

try:
    from Crypto import Random
    from Crypto.Cipher import AES
    from Crypto.Hash import SHA256
    from Crypto.Util import Counter
except ImportError:
    from Cryptodome import Random
    from Cryptodome.Cipher import AES
    from Cryptodome.Hash import SHA256
    from Cryptodome.Util import Counter


def decrypt_attachment(
    ciphertext: bytes | bytearray | memoryview, key: str, hash: str, iv: str, inplace: bool = False
) -> bytes:
    """Decrypt an encrypted attachment.

    Args:
        ciphertext: The data to decrypt.
        key: AES_CTR JWK key object.
        hash: Base64 encoded SHA-256 hash of the ciphertext.
        iv: Base64 encoded 16 byte AES-CTR IV.
        inplace: Should the decryption be performed in-place?
                 The input must be a bytearray or writable memoryview to use this.
    Returns:
        The plaintext bytes.
    Raises:
        EncryptionError: if the integrity check fails.
    """
    expected_hash = unpaddedbase64.decode_base64(hash)

    h = SHA256.new()
    h.update(ciphertext)

    if h.digest() != expected_hash:
        raise DecryptionError("Mismatched SHA-256 digest")

    try:
        byte_key: bytes = unpaddedbase64.decode_base64(key)
    except (binascii.Error, TypeError):
        raise DecryptionError("Error decoding key")

    try:
        byte_iv: bytes = unpaddedbase64.decode_base64(iv)
        if len(byte_iv) != 16:
            raise DecryptionError("Invalid IV length")
        prefix = byte_iv[:8]
        # A non-zero IV counter is not spec-compliant, but some clients still do it,
        # so decode the counter part too.
        initial_value = struct.unpack(">Q", byte_iv[8:])[0]
    except (binascii.Error, TypeError, IndexError, struct.error):
        raise DecryptionError("Error decoding IV")

    ctr = Counter.new(64, prefix=prefix, initial_value=initial_value)

    try:
        cipher = AES.new(byte_key, AES.MODE_CTR, counter=ctr)
    except ValueError as e:
        raise DecryptionError("Failed to create AES cipher") from e

    if inplace:
        cipher.decrypt(ciphertext, ciphertext)
        return ciphertext
    else:
        return cipher.decrypt(ciphertext)


def encrypt_attachment(plaintext: bytes) -> tuple[bytes, EncryptedFile]:
    """Encrypt data in order to send it as an encrypted attachment.

    Args:
        plaintext: The data to encrypt.

    Returns:
        A tuple with the encrypted bytes and a dict containing the info needed
        to decrypt data. See ``encrypted_attachment_generator()`` for the keys.
    """
    values = list(encrypted_attachment_generator(plaintext))
    return b"".join(values[:-1]), values[-1]


def _prepare_encryption() -> tuple[bytes, bytes, AES, SHA256.SHA256Hash]:
    key = Random.new().read(32)
    # 8 bytes IV
    iv = Random.new().read(8)
    # 8 bytes counter, prefixed by the IV
    ctr = Counter.new(64, prefix=iv, initial_value=0)

    cipher = AES.new(key, AES.MODE_CTR, counter=ctr)
    sha256 = SHA256.new()

    return key, iv, cipher, sha256


def inplace_encrypt_attachment(data: bytearray | memoryview) -> EncryptedFile:
    key, iv, cipher, sha256 = _prepare_encryption()

    cipher.encrypt(plaintext=data, output=data)
    sha256.update(data)

    return _get_decryption_info(key, iv, sha256)


def encrypted_attachment_generator(
    data: bytes | Iterable[bytes],
) -> Generator[bytes | EncryptedFile, None, None]:
    """Generator to encrypt data in order to send it as an encrypted
    attachment.

    Unlike ``encrypt_attachment()``, this function lazily encrypts and yields
    data, thus it can be used to encrypt large files without fully loading them
    into memory if an iterable of bytes is passed as data.

    Args:
        data: The data to encrypt.

    Yields:
        The encrypted bytes for each chunk of data.
        The last yielded value will be a dict containing the info needed to decrypt data.
    """

    key, iv, cipher, sha256 = _prepare_encryption()

    if isinstance(data, bytes):
        data = [data]

    for chunk in data:
        encrypted_chunk = cipher.encrypt(chunk)  # in executor
        sha256.update(encrypted_chunk)  # in executor
        yield encrypted_chunk

    yield _get_decryption_info(key, iv, sha256)


def _get_decryption_info(key: bytes, iv: bytes, sha256: SHA256.SHA256Hash) -> EncryptedFile:
    return EncryptedFile(
        version="v2",
        iv=unpaddedbase64.encode_base64(iv + b"\x00" * 8),
        hashes={"sha256": unpaddedbase64.encode_base64(sha256.digest())},
        key=JSONWebKey(
            key_type="oct",
            algorithm="A256CTR",
            extractable=True,
            key_ops=["encrypt", "decrypt"],
            key=unpaddedbase64.encode_base64(key, urlsafe=True),
        ),
    )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Iterable
from pathlib import Path
import asyncio
import json
import logging
import mimetypes
import os
import shutil
import tempfile

try:
    from . import magic
except ImportError:
    magic = None


def _abswhich(program: str) -> str | None:
    path = shutil.which(program)
    return os.path.abspath(path) if path else None


class ConverterError(ChildProcessError):
    pass


class NotInstalledError(ConverterError):
    def __init__(self) -> None:
        super().__init__("failed to transcode media: ffmpeg is not installed")


ffmpeg_path = _abswhich("ffmpeg")
ffmpeg_default_params = ("-hide_banner", "-loglevel", "warning", "-y")

ffprobe_path = _abswhich("ffprobe")
ffprobe_default_params = (
    "-loglevel",
    "quiet",
    "-print_format",
    "json",
    "-show_optional_fields",
    "1",
    "-show_format",
    "-show_streams",
)


async def probe_path(
    input_file: os.PathLike[str] | str,
    logger: logging.Logger | None = None,
) -> Any:
    """
    Probes a media file on the disk using ffprobe.

    Args:
        input_file: The full path to the file.

    Returns:
        A Python object containing the parsed JSON response from ffprobe

    Raises:
        ConverterError: if ffprobe returns a non-zero exit code.
    """
    if ffprobe_path is None:
        raise NotInstalledError()

    input_file = Path(input_file)
    proc = await asyncio.create_subprocess_exec(
        ffprobe_path,
        *ffprobe_default_params,
        str(input_file),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        err_text = stderr.decode("utf-8") if stderr else f"unknown ({proc.returncode})"
        raise ConverterError(f"ffprobe error: {err_text}")
    elif stderr and logger:
        logger.warning(f"ffprobe warning: {stderr.decode('utf-8')}")
    return json.loads(stdout)


async def probe_bytes(
    data: bytes,
    input_mime: str | None = None,
    logger: logging.Logger | None = None,
) -> Any:
    """
    Probe media file data using ffprobe.

    Args:
        data: The bytes of the file to probe.
        input_mime: The mime type of the input data. If not specified, will be guessed using magic.

    Returns:
        A Python object containing the parsed JSON response from ffprobe

    Raises:
        ConverterError: if ffprobe returns a non-zero exit code.
    """
    if ffprobe_path is None:
        raise NotInstalledError()

    if input_mime is None:
        if magic is None:
            raise ValueError("input_mime was not specified and magic is not installed")
        input_mime = magic.mimetype(data)
    input_extension = mimetypes.guess_extension(input_mime)
    with tempfile.TemporaryDirectory(prefix="mautrix_ffmpeg_") as tmpdir:
        input_file = Path(tmpdir) / f"data{input_extension}"
        with open(input_file, "wb") as file:
            file.write(data)
        return await probe_path(input_file=input_file, logger=logger)


async def convert_path(
    input_file: os.PathLike[str] | str,
    output_extension: str | None,
    input_args: Iterable[str] | None = None,
    output_args: Iterable[str] | None = None,
    remove_input: bool = False,
    output_path_override: os.PathLike[str] | str | None = None,
    logger: logging.Logger | None = None,
) -> Path | bytes:
    """
    Convert a media file on the disk using ffmpeg.

    Args:
        input_file: The full path to the file.
        output_extension: The extension that the output file should be.
        input_args: Arguments to tell ffmpeg how to parse the input file.
        output_args: Arguments to tell ffmpeg how to convert the file to reach the wanted output.
        remove_input: Whether the input file should be removed after converting.
                      Not compatible with ``output_path_override``.
        output_path_override: A custom output path to use
                              (instead of using the input path with a different extension).

    Returns:
        The path to the converted file, or the stdout if ``output_path_override`` was set to ``-``.

    Raises:
        ConverterError: if ffmpeg returns a non-zero exit code.
    """
    if ffmpeg_path is None:
        raise NotInstalledError()

    if output_path_override:
        output_file = output_path_override
        if remove_input:
            raise ValueError("remove_input can't be specified with output_path_override")
    elif not output_extension:
        raise ValueError("output_extension or output_path_override is required")
    else:
        input_file = Path(input_file)
        output_file = input_file.parent / f"{input_file.stem}{output_extension}"
    if input_file == output_file:
        output_file = Path(output_file)
        output_file = output_file.parent / f"{output_file.stem}-new{output_extension}"

    proc = await asyncio.create_subprocess_exec(
        ffmpeg_path,
        *ffmpeg_default_params,
        *(input_args or ()),
        "-i",
        str(input_file),
        *(output_args or ()),
        str(output_file),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate()
    if proc.returncode != 0:
        err_text = stderr.decode("utf-8") if stderr else f"unknown ({proc.returncode})"
        raise ConverterError(f"ffmpeg error: {err_text}")
    elif stderr and logger:
        logger.warning(f"ffmpeg warning: {stderr.decode('utf-8')}")
    if remove_input and isinstance(input_file, Path):
        input_file.unlink(missing_ok=True)
    return stdout if output_file == "-" else output_file


async def convert_bytes(
    data: bytes,
    output_extension: str,
    input_args: Iterable[str] | None = None,
    output_args: Iterable[str] | None = None,
    input_mime: str | None = None,
    logger: logging.Logger | None = None,
) -> bytes:
    """
    Convert media file data using ffmpeg.

    Args:
        data: The bytes of the file to convert.
        output_extension: The extension that the output file should be.
        input_args: Arguments to tell ffmpeg how to parse the input file.
        output_args: Arguments to tell ffmpeg how to convert the file to reach the wanted output.
        input_mime: The mime type of the input data. If not specified, will be guessed using magic.

    Returns:
        The converted file as bytes.

    Raises:
        ConverterError: if ffmpeg returns a non-zero exit code.
    """
    if ffmpeg_path is None:
        raise NotInstalledError()

    if input_mime is None:
        if magic is None:
            raise ValueError("input_mime was not specified and magic is not installed")
        input_mime = magic.mimetype(data)
    input_extension = mimetypes.guess_extension(input_mime)
    with tempfile.TemporaryDirectory(prefix="mautrix_ffmpeg_") as tmpdir:
        input_file = Path(tmpdir) / f"data{input_extension}"
        with open(input_file, "wb") as file:
            file.write(data)
        output_file = await convert_path(
            input_file=input_file,
            output_extension=output_extension,
            input_args=input_args,
            output_args=output_args,
            logger=logger,
        )
        with open(output_file, "rb") as file:
            return file.read()


__all__ = [
    "ffmpeg_path",
    "ffmpeg_default_params",
    "ConverterError",
    "NotInstalledError",
    "convert_bytes",
    "convert_path",
    "probe_bytes",
    "probe_path",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, ClassVar
from abc import ABC
import json
import os
import re
import secrets
import time

from mautrix.util.config import (
    BaseFileConfig,
    BaseValidatableConfig,
    ConfigUpdateHelper,
    ForbiddenDefault,
    yaml,
)


class BaseBridgeConfig(BaseFileConfig, BaseValidatableConfig, ABC):
    env_prefix: str | None = None
    registration_path: str
    _registration: dict | None
    _check_tokens: bool
    env: dict[str, Any]

    def __init__(
        self, path: str, registration_path: str, base_path: str, env_prefix: str | None = None
    ) -> None:
        super().__init__(path, base_path)
        self.registration_path = registration_path
        self._registration = None
        self._check_tokens = True
        self.env = {}
        if not self.env_prefix:
            self.env_prefix = env_prefix
        if self.env_prefix:
            env_prefix = f"{self.env_prefix}_"
            for key, value in os.environ.items():
                if not key.startswith(env_prefix):
                    continue
                key = key.removeprefix(env_prefix)
                if value.startswith("json::"):
                    value = json.loads(value.removeprefix("json::"))
                self.env[key] = value

    def __getitem__(self, item: str) -> Any:
        if self.env:
            try:
                sanitized_item = item.replace(".", "_").replace("[", "").replace("]", "").upper()
                return self.env[sanitized_item]
            except KeyError:
                pass
        return super().__getitem__(item)

    def save(self) -> None:
        super().save()
        if self._registration and self.registration_path:
            with open(self.registration_path, "w") as stream:
                yaml.dump(self._registration, stream)

    @staticmethod
    def _new_token() -> str:
        return secrets.token_urlsafe(48)

    @property
    def forbidden_defaults(self) -> list[ForbiddenDefault]:
        return [
            ForbiddenDefault("homeserver.address", "https://example.com"),
            ForbiddenDefault("homeserver.address", "https://matrix.example.com"),
            ForbiddenDefault("homeserver.domain", "example.com"),
        ] + (
            [
                ForbiddenDefault(
                    "appservice.as_token",
                    "This value is generated when generating the registration",
                    "Did you forget to generate the registration?",
                ),
                ForbiddenDefault(
                    "appservice.hs_token",
                    "This value is generated when generating the registration",
                    "Did you forget to generate the registration?",
                ),
            ]
            if self._check_tokens
            else []
        )

    def do_update(self, helper: ConfigUpdateHelper) -> None:
        copy, copy_dict = helper.copy, helper.copy_dict

        copy("homeserver.address")
        copy("homeserver.domain")
        copy("homeserver.verify_ssl")
        copy("homeserver.http_retry_count")
        copy("homeserver.connection_limit")
        copy("homeserver.status_endpoint")
        copy("homeserver.message_send_checkpoint_endpoint")
        copy("homeserver.async_media")
        if self.get("homeserver.asmux", False):
            helper.base["homeserver.software"] = "asmux"
        else:
            copy("homeserver.software")

        copy("appservice.address")
        copy("appservice.hostname")
        copy("appservice.port")
        copy("appservice.max_body_size")

        copy("appservice.tls_cert")
        copy("appservice.tls_key")

        if "appservice.database" in self and self["appservice.database"].startswith("sqlite:///"):
            helper.base["appservice.database"] = self["appservice.database"].replace(
                "sqlite:///", "sqlite:"
            )
        else:
            copy("appservice.database")
        copy("appservice.database_opts")

        copy("appservice.id")
        copy("appservice.bot_username")
        copy("appservice.bot_displayname")
        copy("appservice.bot_avatar")

        copy("appservice.as_token")
        copy("appservice.hs_token")

        copy("appservice.ephemeral_events")

        copy("bridge.management_room_text.welcome")
        copy("bridge.management_room_text.welcome_connected")
        copy("bridge.management_room_text.welcome_unconnected")
        copy("bridge.management_room_text.additional_help")
        copy("bridge.management_room_multiple_messages")

        copy("bridge.encryption.allow")
        copy("bridge.encryption.default")
        copy("bridge.encryption.require")
        copy("bridge.encryption.appservice")
        copy("bridge.encryption.delete_keys.delete_outbound_on_ack")
        copy("bridge.encryption.delete_keys.dont_store_outbound")
        copy("bridge.encryption.delete_keys.ratchet_on_decrypt")
        copy("bridge.encryption.delete_keys.delete_fully_used_on_decrypt")
        copy("bridge.encryption.delete_keys.delete_prev_on_new_session")
        copy("bridge.encryption.delete_keys.delete_on_device_delete")
        copy("bridge.encryption.delete_keys.periodically_delete_expired")
        copy("bridge.encryption.verification_levels.receive")
        copy("bridge.encryption.verification_levels.send")
        copy("bridge.encryption.verification_levels.share")
        copy("bridge.encryption.allow_key_sharing")
        if self.get("bridge.encryption.key_sharing.allow", False):
            helper.base["bridge.encryption.allow_key_sharing"] = True
            require_verif = self.get("bridge.encryption.key_sharing.require_verification", True)
            require_cs = self.get("bridge.encryption.key_sharing.require_cross_signing", False)
            if require_verif:
                helper.base["bridge.encryption.verification_levels.share"] = "verified"
            elif not require_cs:
                helper.base["bridge.encryption.verification_levels.share"] = "unverified"
            # else: default (cross-signed-tofu)
        copy("bridge.encryption.rotation.enable_custom")
        copy("bridge.encryption.rotation.milliseconds")
        copy("bridge.encryption.rotation.messages")
        copy("bridge.encryption.rotation.disable_device_change_key_rotation")

        copy("bridge.relay.enabled")
        copy_dict("bridge.relay.message_formats", override_existing_map=False)

        copy("manhole.enabled")
        copy("manhole.path")
        copy("manhole.whitelist")

        copy("logging")

    @property
    def namespaces(self) -> dict[str, list[dict[str, Any]]]:
        """
        Generate the user ID and room alias namespace config for the registration as specified in
        https://matrix.org/docs/spec/application_service/r0.1.0.html#application-services
        """
        homeserver = self["homeserver.domain"]
        regex_ph = f"regexplaceholder{int(time.time())}"
        username_format = self["bridge.username_template"].format(userid=regex_ph)
        alias_format = (
            self["bridge.alias_template"].format(groupname=regex_ph)
            if "bridge.alias_template" in self
            else None
        )

        return {
            "users": [
                {
                    "exclusive": True,
                    "regex": re.escape(f"@{username_format}:{homeserver}").replace(regex_ph, ".*"),
                }
            ],
            "aliases": [
                {
                    "exclusive": True,
                    "regex": re.escape(f"#{alias_format}:{homeserver}").replace(regex_ph, ".*"),
                }
            ]
            if alias_format
            else [],
        }

    def generate_registration(self) -> None:
        self["appservice.as_token"] = self._new_token()
        self["appservice.hs_token"] = self._new_token()

        namespaces = self.namespaces
        bot_username = self["appservice.bot_username"]
        homeserver_domain = self["homeserver.domain"]
        namespaces.setdefault("users", []).append(
            {
                "exclusive": True,
                "regex": re.escape(f"@{bot_username}:{homeserver_domain}"),
            }
        )

        self._registration = {
            "id": self["appservice.id"],
            "as_token": self["appservice.as_token"],
            "hs_token": self["appservice.hs_token"],
            "namespaces": namespaces,
            "url": self["appservice.address"],
            "sender_localpart": self._new_token(),
            "rate_limited": False,
        }

        if self["appservice.ephemeral_events"]:
            self._registration["de.sorunome.msc2409.push_ephemeral"] = True
            self._registration["push_ephemeral"] = True



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import asyncio
import logging


class SimpleLock:
    _event: asyncio.Event
    log: logging.Logger | None
    message: str | None
    noop_mode: bool

    def __init__(
        self,
        message: str | None = None,
        log: logging.Logger | None = None,
        noop_mode: bool = False,
    ) -> None:
        self.noop_mode = noop_mode
        if not noop_mode:
            self._event = asyncio.Event()
            self._event.set()
        self.log = log
        self.message = message

    def __enter__(self) -> None:
        if not self.noop_mode:
            self._event.clear()

    async def __aenter__(self) -> None:
        self.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not self.noop_mode:
            self._event.set()

    def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        self.__exit__(exc_type, exc_val, exc_tb)

    @property
    def locked(self) -> bool:
        return not self.noop_mode and not self._event.is_set()

    async def wait(self, task: str | None = None) -> None:
        if self.locked:
            if self.log and self.message:
                self.log.debug(self.message, task)
            await self._event.wait()



from mautrix import __optional_imports__

from .connection import LoggingConnection as Connection
from .database import Database
from .errors import (
    DatabaseException,
    DatabaseNotOwned,
    ForeignTablesFound,
    UnsupportedDatabaseVersion,
)
from .scheme import Scheme
from .upgrade import UpgradeTable, register_upgrade

try:
    from .asyncpg import PostgresDatabase
except ImportError:
    if __optional_imports__:
        raise
    PostgresDatabase = None

try:
    from aiosqlite import Cursor as SQLiteCursor

    from .aiosqlite import SQLiteDatabase
except ImportError:
    if __optional_imports__:
        raise
    SQLiteDatabase = None
    SQLiteCursor = None

__all__ = [
    "Database",
    "UpgradeTable",
    "register_upgrade",
    "PostgresDatabase",
    "SQLiteDatabase",
    "SQLiteCursor",
    "Connection",
    "Scheme",
    "DatabaseException",
    "DatabaseNotOwned",
    "UnsupportedDatabaseVersion",
    "ForeignTablesFound",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, AsyncIterable, Literal
from contextlib import contextmanager
import asyncio
import time

from mautrix import __optional_imports__
from mautrix.api import MediaPath, Method
from mautrix.errors import MatrixResponseError, make_request_error
from mautrix.types import (
    ContentURI,
    MediaCreateResponse,
    MediaRepoConfig,
    MXOpenGraph,
    SerializerError,
)
from mautrix.util import background_task
from mautrix.util.async_body import async_iter_bytes
from mautrix.util.opt_prometheus import Histogram

from ..base import BaseClientAPI

try:
    from mautrix.util import magic
except ImportError:
    if __optional_imports__:
        raise
    magic = None  # type: ignore

UPLOAD_TIME = Histogram(
    "bridge_media_upload_time",
    "Time spent uploading media (milliseconds per megabyte)",
    buckets=[10, 25, 50, 100, 250, 500, 750, 1000, 2500, 5000, 10000],
)


class MediaRepositoryMethods(BaseClientAPI):
    """
    Methods in section 13.8 Content Repository of the spec. These methods are used for uploading and
    downloading content from the media repository and for getting URL previews without leaking
    client IPs.

    See also: `API reference <https://spec.matrix.org/v1.7/client-server-api/#content-repository>`__
    """

    async def create_mxc(self) -> MediaCreateResponse:
        """
        Create a media ID for uploading media to the homeserver.

        See also: `API reference <https://spec.matrix.org/unstable/client-server-api/#post_matrixmediav1create>`__

        Returns:
            MediaCreateResponse Containing the MXC URI that can be used to upload a file to later
        """
        resp = await self.api.request(Method.POST, MediaPath.v1.create)
        return MediaCreateResponse.deserialize(resp)

    @contextmanager
    def _observe_upload_time(self, size: int | None, mxc: ContentURI | None = None) -> None:
        start = time.monotonic_ns()
        yield
        duration = time.monotonic_ns() - start
        if mxc:
            duration_sec = duration / 1000**3
            self.log.debug(f"Completed asynchronous upload of {mxc} in {duration_sec:.3f} seconds")
        if size:
            UPLOAD_TIME.observe(duration / size)

    async def upload_media(
        self,
        data: bytes | bytearray | AsyncIterable[bytes],
        mime_type: str | None = None,
        filename: str | None = None,
        size: int | None = None,
        mxc: ContentURI | None = None,
        async_upload: bool = False,
    ) -> ContentURI:
        """
        Upload a file to the content repository.

        See also: `API reference <https://spec.matrix.org/v1.7/client-server-api/#post_matrixmediav3upload>`__

        Args:
            data: The data to upload.
            mime_type: The MIME type to send with the upload request.
            filename: The filename to send with the upload request.
            size: The file size to send with the upload request.
            mxc: An existing MXC URI which doesn't have content yet to upload into.
            async_upload: Should the media be uploaded in the background?
                If ``True``, this will create a MXC URI using :meth:`create_mxc`, start uploading
                in the background, and then immediately return the created URI. This is mutually
                exclusive with manually passing the ``mxc`` parameter.

        Returns:
            The MXC URI to the uploaded file.

        Raises:
            MatrixResponseError: If the response does not contain a ``content_uri`` field.
            ValueError: if both ``async_upload`` and ``mxc`` are provided at the same time.
        """
        if magic and isinstance(data, bytes):
            mime_type = mime_type or magic.mimetype(data)
        headers = {}
        if mime_type:
            headers["Content-Type"] = mime_type
        if size:
            headers["Content-Length"] = str(size)
        elif isinstance(data, (bytes, bytearray)):
            size = len(data)
        query = {}
        if filename:
            query["filename"] = filename

        upload_url = None

        if async_upload:
            if mxc:
                raise ValueError("async_upload and mxc can't be provided simultaneously")
            create_response = await self.create_mxc()
            mxc = create_response.content_uri
            upload_url = create_response.unstable_upload_url

        path = MediaPath.v3.upload
        method = Method.POST
        if mxc:
            server_name, media_id = self.api.parse_mxc_uri(mxc)
            if upload_url is None:
                path = MediaPath.v3.upload[server_name][media_id]
                method = Method.PUT
            else:
                path = (
                    MediaPath.unstable["com.beeper.msc3870"].upload[server_name][media_id].complete
                )

        if upload_url is not None:
            task = self._upload_to_url(upload_url, path, headers, data, post_upload_query=query)
        else:
            task = self.api.request(
                method, path, content=data, headers=headers, query_params=query
            )

        if async_upload:

            async def _try_upload():
                try:
                    with self._observe_upload_time(size, mxc):
                        await task
                except Exception as e:
                    self.log.error(f"Failed to upload {mxc}: {type(e).__name__}: {e}")

            background_task.create(_try_upload())
            return mxc
        else:
            with self._observe_upload_time(size):
                resp = await task
            try:
                return resp["content_uri"]
            except KeyError:
                raise MatrixResponseError("`content_uri` not in response.")

    async def download_media(self, url: ContentURI, timeout_ms: int | None = None) -> bytes:
        """
        Download a file from the content repository.

        See also: `API reference <https://spec.matrix.org/v1.7/client-server-api/#get_matrixmediav3downloadservernamemediaid>`__

        Args:
            url: The MXC URI to download.
            timeout_ms: The maximum number of milliseconds that the client is willing to wait to
                start receiving data. Used for asynchronous uploads.

        Returns:
            The raw downloaded data.
        """
        url = self.api.get_download_url(url)
        query_params: dict[str, Any] = {"allow_redirect": "true"}
        if timeout_ms is not None:
            query_params["timeout_ms"] = timeout_ms
        async with self.api.session.get(url, params=query_params) as response:
            return await response.read()

    async def download_thumbnail(
        self,
        url: ContentURI,
        width: int | None = None,
        height: int | None = None,
        resize_method: Literal["crop", "scale"] = None,
        allow_remote: bool = True,
        timeout_ms: int | None = None,
    ):
        """
        Download a thumbnail for a file in the content repository.

        See also: `API reference <https://spec.matrix.org/v1.7/client-server-api/#get_matrixmediav3thumbnailservernamemediaid>`__

        Args:
            url: The MXC URI to download.
            width: The _desired_ width of the thumbnail. The actual thumbnail may not match the size
                specified.
            height: The _desired_ height of the thumbnail. The actual thumbnail may not match the
                size specified.
            resize_method: The desired resizing method. Either ``crop`` or ``scale``.
            allow_remote: Indicates to the server that it should not attempt to fetch the media if
                it is deemed remote. This is to prevent routing loops where the server contacts
                itself.
            timeout_ms: The maximum number of milliseconds that the client is willing to wait to
                start receiving data. Used for asynchronous Uploads.

        Returns:
            The raw downloaded data.
        """
        url = self.api.get_download_url(url, download_type="thumbnail")
        query_params: dict[str, Any] = {"allow_redirect": "true"}
        if width is not None:
            query_params["width"] = width
        if height is not None:
            query_params["height"] = height
        if resize_method is not None:
            query_params["method"] = resize_method
        if allow_remote is not None:
            query_params["allow_remote"] = allow_remote
        if timeout_ms is not None:
            query_params["timeout_ms"] = timeout_ms
        async with self.api.session.get(url, params=query_params) as response:
            return await response.read()

    async def get_url_preview(self, url: str, timestamp: int | None = None) -> MXOpenGraph:
        """
        Get information about a URL for a client.

        See also: `API reference <https://spec.matrix.org/v1.2/client-server-api/#get_matrixmediav3preview_url>`__

        Args:
            url: The URL to get a preview of.
            timestamp: The preferred point in time to return a preview for. The server may return a
                newer version if it does not have the requested version available.
        """
        query_params = {"url": url}
        if timestamp is not None:
            query_params["ts"] = timestamp
        content = await self.api.request(
            Method.GET, MediaPath.v3.preview_url, query_params=query_params
        )
        try:
            return MXOpenGraph.deserialize(content)
        except SerializerError as e:
            raise MatrixResponseError("Invalid MXOpenGraph in response.") from e

    async def get_media_repo_config(self) -> MediaRepoConfig:
        """
        This endpoint allows clients to retrieve the configuration of the content repository, such
        as upload limitations. Clients SHOULD use this as a guide when using content repository
        endpoints. All values are intentionally left optional. Clients SHOULD follow the advice
        given in the field description when the field is not available.

        **NOTE:** Both clients and server administrators should be aware that proxies between the
        client and the server may affect the apparent behaviour of content repository APIs, for
        example, proxies may enforce a lower upload size limit than is advertised by the server on
        this endpoint.

        See also: `API reference <https://matrix.org/docs/spec/client_server/r0.4.0.html#get-matrix-media-r0-config>`__

        Returns:
            The media repository config.
        """
        content = await self.api.request(Method.GET, MediaPath.v3.config)
        try:
            return MediaRepoConfig.deserialize(content)
        except SerializerError as e:
            raise MatrixResponseError("Invalid MediaRepoConfig in response") from e

    async def _upload_to_url(
        self,
        upload_url: str,
        post_upload_path: str,
        headers: dict[str, str],
        data: bytes | bytearray | AsyncIterable[bytes],
        post_upload_query: dict[str, str],
        min_iter_size: int = 25 * 1024 * 1024,
    ) -> None:
        retry_count = self.api.default_retry_count
        backoff = 2
        do_fake_iter = data and hasattr(data, "__len__") and len(data) > min_iter_size
        if do_fake_iter:
            headers["Content-Length"] = str(len(data))
        while True:
            self.log.debug("Uploading media to external URL %s", upload_url)
            upload_response = None
            try:
                req_data = async_iter_bytes(data) if do_fake_iter else data
                upload_response = await self.api.session.put(
                    upload_url, data=req_data, headers=headers
                )
                upload_response.raise_for_status()
            except Exception as e:
                if retry_count <= 0:
                    raise make_request_error(
                        http_status=upload_response.status if upload_response else -1,
                        text=(await upload_response.text()) if upload_response else "",
                        errcode="COM.BEEPER.EXTERNAL_UPLOAD_ERROR",
                        message=None,
                    )
                self.log.warning(
                    f"Uploading media to external URL {upload_url} failed: {e}, "
                    f"retrying in {backoff} seconds",
                )
                await asyncio.sleep(backoff)
                backoff *= 2
                retry_count -= 1
            else:
                break

        await self.api.request(Method.POST, post_upload_path, query_params=post_upload_query)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import NewType, Union

from ..primitive import JSON
from ..util import Obj, deserializer
from .account_data import AccountDataEvent, AccountDataEventContent
from .base import EventType, GenericEvent
from .beeper import BeeperMessageStatusEvent, BeeperMessageStatusEventContent
from .encrypted import EncryptedEvent, EncryptedEventContent
from .ephemeral import (
    EphemeralEvent,
    PresenceEvent,
    ReceiptEvent,
    ReceiptEventContent,
    TypingEvent,
    TypingEventContent,
)
from .message import MessageEvent, MessageEventContent
from .reaction import ReactionEvent, ReactionEventContent
from .redaction import RedactionEvent, RedactionEventContent
from .state import StateEvent, StateEventContent
from .to_device import ASToDeviceEvent, ToDeviceEvent, ToDeviceEventContent
from .voip import CallEvent, CallEventContent, type_to_class as voip_types

Event = NewType(
    "Event",
    Union[
        MessageEvent,
        ReactionEvent,
        RedactionEvent,
        StateEvent,
        TypingEvent,
        ReceiptEvent,
        PresenceEvent,
        EncryptedEvent,
        ToDeviceEvent,
        ASToDeviceEvent,
        CallEvent,
        BeeperMessageStatusEvent,
        GenericEvent,
    ],
)

EventContent = Union[
    MessageEventContent,
    RedactionEventContent,
    ReactionEventContent,
    StateEventContent,
    AccountDataEventContent,
    ReceiptEventContent,
    TypingEventContent,
    EncryptedEventContent,
    ToDeviceEventContent,
    CallEventContent,
    BeeperMessageStatusEventContent,
    Obj,
]


@deserializer(Event)
def deserialize_event(data: JSON) -> Event:
    event_type = EventType.find(data.get("type", None))
    if event_type == EventType.ROOM_MESSAGE:
        return MessageEvent.deserialize(data)
    elif event_type == EventType.STICKER:
        data.get("content", {})["msgtype"] = "m.sticker"
        return MessageEvent.deserialize(data)
    elif event_type == EventType.REACTION:
        return ReactionEvent.deserialize(data)
    elif event_type == EventType.ROOM_REDACTION:
        return RedactionEvent.deserialize(data)
    elif event_type == EventType.ROOM_ENCRYPTED:
        return EncryptedEvent.deserialize(data)
    elif event_type in voip_types.keys():
        return CallEvent.deserialize(data, event_type=event_type)
    elif event_type.is_to_device:
        return ToDeviceEvent.deserialize(data)
    elif event_type.is_state:
        return StateEvent.deserialize(data)
    elif event_type.is_account_data:
        return AccountDataEvent.deserialize(data)
    elif event_type.is_ephemeral:
        return EphemeralEvent.deserialize(data)
    elif event_type == EventType.BEEPER_MESSAGE_STATUS:
        return BeeperMessageStatusEvent.deserialize(data)
    else:
        return GenericEvent.deserialize(data)


setattr(Event, "deserialize", deserialize_event)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Dict, Optional, Tuple
from abc import ABC
import time

from mautrix.client.state_store import StateStore as ClientStateStore
from mautrix.types import EventID, RoomID, UserID


class ASStateStore(ClientStateStore, ABC):
    _presence: Dict[UserID, str]
    _read: Dict[Tuple[RoomID, UserID], EventID]
    _registered: Dict[UserID, bool]

    def __init__(self) -> None:
        self._registered = {}
        # Non-persistent storage
        self._presence = {}
        self._read = {}

    async def is_registered(self, user_id: UserID) -> bool:
        """
        Check if a given user is registered.

        This should always return ``True`` for double puppets, because they're always registered
        beforehand and shouldn't be attempted to register by the bridge.

        Args:
            user_id: The user ID to check.

        Returns:
            ``True`` if the user is registered, ``False`` otherwise.
        """
        if not user_id:
            raise ValueError("user_id is empty")
        return self._registered.get(user_id, False)

    async def registered(self, user_id: UserID) -> None:
        """
        Mark the given user ID as registered.

        Args:
            user_id: The user ID to mark as registered.
        """
        if not user_id:
            raise ValueError("user_id is empty")
        self._registered[user_id] = True

    def set_presence(self, user_id: UserID, presence: str) -> None:
        self._presence[user_id] = presence

    def has_presence(self, user_id: UserID, presence: str) -> bool:
        try:
            return self._presence[user_id] == presence
        except KeyError:
            return False

    def set_read(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        self._read[(room_id, user_id)] = event_id

    def get_read(self, room_id: RoomID, user_id: UserID) -> Optional[EventID]:
        try:
            return self._read[(room_id, user_id)]
        except KeyError:
            return None



# Copyright © 2018, 2019 Damir Jelić <poljar@termina.org.uk>
# Copyright © 2019 miruka <miruka@disroot.org>
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted, provided that the
# above copyright notice and this permission notice appear in all copies.
#
# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import AsyncGenerator, AsyncIterable, Iterable
from functools import partial
import asyncio
import io

from mautrix.types import EncryptedFile

from .attachments import _get_decryption_info, _prepare_encryption, inplace_encrypt_attachment


async def async_encrypt_attachment(
    data: bytes | Iterable[bytes] | AsyncIterable[bytes] | io.BufferedIOBase,
) -> AsyncGenerator[bytes | EncryptedFile, None]:
    """Async generator to encrypt data in order to send it as an encrypted
    attachment.

    This function lazily encrypts and yields data, thus it can be used to
    encrypt large files without fully loading them into memory if an iterable
    or async iterable of bytes is passed as data.

    Args:
        data: The data to encrypt.
            Passing an async iterable allows the file data to be read in an asynchronous and lazy
            (without reading the entire file into memory) way.
            Passing a non-async iterable or standard open binary file object will still allow the
            data to be read lazily, but not asynchronously.

    Yields:
        The encrypted bytes for each chunk of data.
        The last yielded value will be a dict containing the info needed to
        decrypt data. The keys are:
        | key: AES-CTR JWK key object.
        | iv: Base64 encoded 16 byte AES-CTR IV.
        | hashes.sha256: Base64 encoded SHA-256 hash of the ciphertext.
    """

    key, iv, cipher, sha256 = _prepare_encryption()

    loop = asyncio.get_running_loop()

    async for chunk in async_generator_from_data(data):
        update_crypt = partial(cipher.encrypt, chunk)
        crypt_chunk = await loop.run_in_executor(None, update_crypt)

        update_hash = partial(sha256.update, crypt_chunk)
        await loop.run_in_executor(None, update_hash)

        yield crypt_chunk

    yield _get_decryption_info(key, iv, sha256)


async def async_inplace_encrypt_attachment(data: bytearray) -> EncryptedFile:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, partial(inplace_encrypt_attachment, data))


async def async_generator_from_data(
    data: bytes | Iterable[bytes] | AsyncIterable[bytes] | io.BufferedIOBase,
    chunk_size: int = 4 * 1024,
) -> AsyncGenerator[bytes, None]:
    if isinstance(data, bytes):
        chunks = (data[i : i + chunk_size] for i in range(0, len(data), chunk_size))
        for chunk in chunks:
            yield chunk
    elif isinstance(data, io.BufferedIOBase):
        while True:
            chunk = data.read(chunk_size)
            if not chunk:
                return
            yield chunk
    elif isinstance(data, Iterable):
        for chunk in data:
            yield chunk
    elif isinstance(data, AsyncIterable):
        async for chunk in data:
            yield chunk
    else:
        raise TypeError(f"Unknown type for data: {data!r}")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable
from abc import ABC, abstractmethod

from mautrix.types import (
    EventType,
    Member,
    Membership,
    MemberStateEventContent,
    PowerLevelStateEventContent,
    RoomEncryptionStateEventContent,
    RoomID,
    StateEvent,
    UserID,
)


class StateStore(ABC):
    async def open(self) -> None:
        pass

    async def close(self) -> None:
        await self.flush()

    async def flush(self) -> None:
        pass

    @abstractmethod
    async def get_member(self, room_id: RoomID, user_id: UserID) -> Member | None:
        pass

    @abstractmethod
    async def set_member(
        self, room_id: RoomID, user_id: UserID, member: Member | MemberStateEventContent
    ) -> None:
        pass

    @abstractmethod
    async def set_membership(
        self, room_id: RoomID, user_id: UserID, membership: Membership
    ) -> None:
        pass

    @abstractmethod
    async def get_member_profiles(
        self,
        room_id: RoomID,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> dict[UserID, Member]:
        pass

    async def get_members(
        self,
        room_id: RoomID,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> list[UserID]:
        profiles = await self.get_member_profiles(room_id, memberships)
        return list(profiles.keys())

    async def get_members_filtered(
        self,
        room_id: RoomID,
        not_prefix: str,
        not_suffix: str,
        not_id: str,
        memberships: tuple[Membership, ...] = (Membership.JOIN, Membership.INVITE),
    ) -> list[UserID]:
        """
        A filtered version of get_members that only returns user IDs that aren't operated by a
        bridge. This should return the same as :meth:`get_members`, except users where the user ID
        is equal to not_id OR it starts with not_prefix AND ends with not_suffix.

        The default implementation simply calls :meth:`get_members`, but databases can implement
        this more efficiently.

        Args:
            room_id: The room ID to find.
            not_prefix: The user ID prefix to disallow.
            not_suffix: The user ID suffix to disallow.
            not_id: The user ID to disallow.
            memberships: The membership states to include.
        """
        members = await self.get_members(room_id, memberships=memberships)
        return [
            user_id
            for user_id in members
            if user_id != not_id
            and not (user_id.startswith(not_prefix) and user_id.endswith(not_suffix))
        ]

    @abstractmethod
    async def set_members(
        self,
        room_id: RoomID,
        members: dict[UserID, Member | MemberStateEventContent],
        only_membership: Membership | None = None,
    ) -> None:
        pass

    @abstractmethod
    async def has_full_member_list(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def has_power_levels_cached(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def get_power_levels(self, room_id: RoomID) -> PowerLevelStateEventContent | None:
        pass

    @abstractmethod
    async def set_power_levels(
        self, room_id: RoomID, content: PowerLevelStateEventContent
    ) -> None:
        pass

    @abstractmethod
    async def has_encryption_info_cached(self, room_id: RoomID) -> bool:
        pass

    @abstractmethod
    async def is_encrypted(self, room_id: RoomID) -> bool | None:
        pass

    @abstractmethod
    async def get_encryption_info(self, room_id: RoomID) -> RoomEncryptionStateEventContent | None:
        pass

    @abstractmethod
    async def set_encryption_info(
        self, room_id: RoomID, content: RoomEncryptionStateEventContent | dict[str, any]
    ) -> None:
        pass

    async def update_state(self, evt: StateEvent) -> None:
        if evt.type == EventType.ROOM_POWER_LEVELS:
            await self.set_power_levels(evt.room_id, evt.content)
        elif evt.type == EventType.ROOM_MEMBER:
            evt.unsigned["mautrix_prev_membership"] = await self.get_member(
                evt.room_id, UserID(evt.state_key)
            )
            await self.set_member(evt.room_id, UserID(evt.state_key), evt.content)
        elif evt.type == EventType.ROOM_ENCRYPTION:
            await self.set_encryption_info(evt.room_id, evt.content)

    async def get_membership(self, room_id: RoomID, user_id: UserID) -> Membership:
        member = await self.get_member(room_id, user_id)
        return member.membership if member else Membership.LEAVE

    async def is_joined(self, room_id: RoomID, user_id: UserID) -> bool:
        return (await self.get_membership(room_id, user_id)) == Membership.JOIN

    def joined(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.JOIN)

    def invited(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.INVITE)

    def left(self, room_id: RoomID, user_id: UserID) -> Awaitable[None]:
        return self.set_membership(room_id, user_id, Membership.LEAVE)

    async def has_power_level(
        self, room_id: RoomID, user_id: UserID, event_type: EventType
    ) -> bool | None:
        room_levels = await self.get_power_levels(room_id)
        if not room_levels:
            return None
        return room_levels.get_user_level(user_id) >= room_levels.get_event_level(event_type)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Awaitable, Callable, Type, TypeVar
from abc import ABC, abstractmethod
from contextlib import suppress
from enum import Enum, Flag, auto
import asyncio
import time

from mautrix.errors import MUnknownToken
from mautrix.types import (
    JSON,
    AccountDataEvent,
    BaseMessageEventContentFuncs,
    DeviceLists,
    DeviceOTKCount,
    EphemeralEvent,
    Event,
    EventType,
    Filter,
    FilterID,
    GenericEvent,
    MessageEvent,
    PresenceState,
    SerializerError,
    StateEvent,
    StrippedStateEvent,
    SyncToken,
    ToDeviceEvent,
    UserID,
)
from mautrix.util.logging import TraceLogger

from . import dispatcher
from .state_store import MemorySyncStore, SyncStore

EventHandler = Callable[[Event], Awaitable[None]]

T = TypeVar("T", bound=Event)


class SyncStream(Flag):
    INTERNAL = auto()

    JOINED_ROOM = auto()
    INVITED_ROOM = auto()
    LEFT_ROOM = auto()

    TIMELINE = auto()
    STATE = auto()
    EPHEMERAL = auto()
    ACCOUNT_DATA = auto()
    TO_DEVICE = auto()


class InternalEventType(Enum):
    SYNC_STARTED = auto()
    SYNC_ERRORED = auto()
    SYNC_SUCCESSFUL = auto()
    SYNC_STOPPED = auto()

    JOIN = auto()
    PROFILE_CHANGE = auto()
    INVITE = auto()
    REJECT_INVITE = auto()
    DISINVITE = auto()
    LEAVE = auto()
    KICK = auto()
    BAN = auto()
    UNBAN = auto()

    DEVICE_LISTS = auto()
    DEVICE_OTK_COUNT = auto()


class Syncer(ABC):
    loop: asyncio.AbstractEventLoop
    log: TraceLogger
    mxid: UserID

    global_event_handlers: list[tuple[EventHandler, bool]]
    event_handlers: dict[EventType | InternalEventType, list[tuple[EventHandler, bool]]]
    dispatchers: dict[Type[dispatcher.Dispatcher], dispatcher.Dispatcher]
    syncing_task: asyncio.Task | None
    ignore_initial_sync: bool
    ignore_first_sync: bool
    presence: PresenceState

    sync_store: SyncStore

    def __init__(self, sync_store: SyncStore) -> None:
        self.global_event_handlers = []
        self.event_handlers = {}
        self.dispatchers = {}
        self.syncing_task = None
        self.ignore_initial_sync = False
        self.ignore_first_sync = False
        self.presence = PresenceState.ONLINE

        self.sync_store = sync_store or MemorySyncStore()

    def on(
        self, var: EventHandler | EventType | InternalEventType
    ) -> EventHandler | Callable[[EventHandler], EventHandler]:
        """
        Add a new event handler. This method is for decorator usage.
        Use :meth:`add_event_handler` if you don't use a decorator.

        Args:
            var: Either the handler function or the event type to handle.

        Returns:
            If ``var`` was the handler function, the handler function is returned.

            If ``var`` was an event type, a function that takes the handler function as an argument
            is returned.

        Examples:
            >>> from mautrix.client import Client
            >>> cli = Client(...)
            >>> @cli.on(EventType.ROOM_MESSAGE)
            >>> def handler(event: MessageEvent) -> None:
            ...     pass
        """
        if isinstance(var, (EventType, InternalEventType)):

            def decorator(func: EventHandler) -> EventHandler:
                self.add_event_handler(var, func)
                return func

            return decorator
        else:
            self.add_event_handler(EventType.ALL, var)
            return var

    def add_dispatcher(self, dispatcher_type: Type[dispatcher.Dispatcher]) -> None:
        if dispatcher_type in self.dispatchers:
            return
        self.log.debug(f"Enabling {dispatcher_type.__name__}")
        self.dispatchers[dispatcher_type] = dispatcher_type(self)
        self.dispatchers[dispatcher_type].register()

    def remove_dispatcher(self, dispatcher_type: Type[dispatcher.Dispatcher]) -> None:
        if dispatcher_type not in self.dispatchers:
            return
        self.log.debug(f"Disabling {dispatcher_type.__name__}")
        self.dispatchers[dispatcher_type].unregister()
        del self.dispatchers[dispatcher_type]

    def add_event_handler(
        self,
        event_type: InternalEventType | EventType,
        handler: EventHandler,
        wait_sync: bool = False,
    ) -> None:
        """
        Add a new event handler.

        Args:
            event_type: The event type to add. If not specified, the handler will be called for all
                event types.
            handler: The handler function to add.
            wait_sync: Whether or not the handler should be awaited before the next sync request.
        """
        if not isinstance(event_type, (EventType, InternalEventType)):
            raise ValueError("Invalid event type")
        if event_type == EventType.ALL:
            self.global_event_handlers.append((handler, wait_sync))
        else:
            self.event_handlers.setdefault(event_type, []).append((handler, wait_sync))

    def remove_event_handler(
        self, event_type: EventType | InternalEventType, handler: EventHandler
    ) -> None:
        """
        Remove an event handler.

        Args:
            handler: The handler function to remove.
            event_type: The event type to remove the handler function from.
        """
        if not isinstance(event_type, (EventType, InternalEventType)):
            raise ValueError("Invalid event type")
        try:
            handler_list = (
                self.global_event_handlers
                if event_type == EventType.ALL
                else self.event_handlers[event_type]
            )
        except KeyError:
            # No handlers for this event type registered
            return

        # FIXME this is a bit hacky
        with suppress(ValueError):
            handler_list.remove((handler, True))
        with suppress(ValueError):
            handler_list.remove((handler, False))

        if len(handler_list) == 0 and event_type != EventType.ALL:
            del self.event_handlers[event_type]

    def dispatch_event(self, event: Event | None, source: SyncStream) -> list[asyncio.Task]:
        """
        Send the given event to all applicable event handlers.

        Args:
            event: The event to send.
            source: The sync stream the event was received in.
        """
        if event is None:
            return []
        if isinstance(event.content, BaseMessageEventContentFuncs):
            event.content.trim_reply_fallback()
        if getattr(event, "state_key", None) is not None:
            event.type = event.type.with_class(EventType.Class.STATE)
        elif source & SyncStream.EPHEMERAL:
            event.type = event.type.with_class(EventType.Class.EPHEMERAL)
        elif source & SyncStream.ACCOUNT_DATA:
            event.type = event.type.with_class(EventType.Class.ACCOUNT_DATA)
        elif source & SyncStream.TO_DEVICE:
            event.type = event.type.with_class(EventType.Class.TO_DEVICE)
        else:
            event.type = event.type.with_class(EventType.Class.MESSAGE)
        setattr(event, "source", source)
        return self.dispatch_manual_event(event.type, event, include_global_handlers=True)

    async def _catch_errors(self, handler: EventHandler, data: Any) -> None:
        try:
            await handler(data)
        except Exception:
            self.log.exception("Failed to run handler")

    def dispatch_manual_event(
        self,
        event_type: EventType | InternalEventType,
        data: Any,
        include_global_handlers: bool = False,
        force_synchronous: bool = False,
    ) -> list[asyncio.Task]:
        handlers = self.event_handlers.get(event_type, [])
        if include_global_handlers:
            handlers = self.global_event_handlers + handlers
        tasks = []
        for handler, wait_sync in handlers:
            task = asyncio.create_task(self._catch_errors(handler, data))
            if force_synchronous or wait_sync:
                tasks.append(task)
        return tasks

    async def run_internal_event(
        self, event_type: InternalEventType, custom_type: Any = None, **kwargs: Any
    ) -> None:
        kwargs["source"] = SyncStream.INTERNAL
        tasks = self.dispatch_manual_event(
            event_type,
            custom_type if custom_type is not None else kwargs,
            include_global_handlers=False,
        )
        await asyncio.gather(*tasks)

    def dispatch_internal_event(
        self, event_type: InternalEventType, custom_type: Any = None, **kwargs: Any
    ) -> list[asyncio.Task]:
        kwargs["source"] = SyncStream.INTERNAL
        return self.dispatch_manual_event(
            event_type,
            custom_type if custom_type is not None else kwargs,
            include_global_handlers=False,
        )

    def _try_deserialize(self, type: Type[T], data: JSON) -> T | GenericEvent:
        try:
            return type.deserialize(data)
        except SerializerError as e:
            self.log.trace("Deserialization error traceback", exc_info=True)
            self.log.warning(f"Failed to deserialize {data} into {type.__name__}: {e}")
            try:
                return GenericEvent.deserialize(data)
            except SerializerError:
                return None

    def handle_sync(self, data: JSON) -> list[asyncio.Task]:
        """
        Handle a /sync object.

        Args:
            data: The data from a /sync request.
        """
        tasks = []

        otk_count = data.get("device_one_time_keys_count", {})
        tasks += self.dispatch_internal_event(
            InternalEventType.DEVICE_OTK_COUNT,
            custom_type=DeviceOTKCount(
                curve25519=otk_count.get("curve25519", 0),
                signed_curve25519=otk_count.get("signed_curve25519", 0),
            ),
        )

        device_lists = data.get("device_lists", {})
        tasks += self.dispatch_internal_event(
            InternalEventType.DEVICE_LISTS,
            custom_type=DeviceLists(
                changed=device_lists.get("changed", []),
                left=device_lists.get("left", []),
            ),
        )

        for raw_event in data.get("account_data", {}).get("events", []):
            tasks += self.dispatch_event(
                self._try_deserialize(AccountDataEvent, raw_event), source=SyncStream.ACCOUNT_DATA
            )
        for raw_event in data.get("ephemeral", {}).get("events", []):
            tasks += self.dispatch_event(
                self._try_deserialize(EphemeralEvent, raw_event), source=SyncStream.EPHEMERAL
            )
        for raw_event in data.get("to_device", {}).get("events", []):
            tasks += self.dispatch_event(
                self._try_deserialize(ToDeviceEvent, raw_event), source=SyncStream.TO_DEVICE
            )

        rooms = data.get("rooms", {})
        for room_id, room_data in rooms.get("join", {}).items():
            for raw_event in room_data.get("state", {}).get("events", []):
                raw_event["room_id"] = room_id
                tasks += self.dispatch_event(
                    self._try_deserialize(StateEvent, raw_event),
                    source=SyncStream.JOINED_ROOM | SyncStream.STATE,
                )

            for raw_event in room_data.get("timeline", {}).get("events", []):
                raw_event["room_id"] = room_id
                tasks += self.dispatch_event(
                    self._try_deserialize(Event, raw_event),
                    source=SyncStream.JOINED_ROOM | SyncStream.TIMELINE,
                )

            for raw_event in room_data.get("ephemeral", {}).get("events", []):
                raw_event["room_id"] = room_id
                tasks += self.dispatch_event(
                    self._try_deserialize(EphemeralEvent, raw_event),
                    source=SyncStream.JOINED_ROOM | SyncStream.EPHEMERAL,
                )
        for room_id, room_data in rooms.get("invite", {}).items():
            events: list[dict[str, JSON]] = room_data.get("invite_state", {}).get("events", [])
            for raw_event in events:
                raw_event["room_id"] = room_id
            raw_invite = next(
                raw_event
                for raw_event in events
                if raw_event.get("type", "") == "m.room.member"
                and raw_event.get("state_key", "") == self.mxid
            )
            # These aren't required by the spec, so make sure they're set
            raw_invite.setdefault("event_id", None)
            raw_invite.setdefault("origin_server_ts", int(time.time() * 1000))

            invite = self._try_deserialize(StateEvent, raw_invite)
            invite.unsigned.invite_room_state = [
                self._try_deserialize(StrippedStateEvent, raw_event)
                for raw_event in events
                if raw_event != raw_invite
            ]
            tasks += self.dispatch_event(invite, source=SyncStream.INVITED_ROOM | SyncStream.STATE)
        for room_id, room_data in rooms.get("leave", {}).items():
            for raw_event in room_data.get("timeline", {}).get("events", []):
                if "state_key" in raw_event:
                    raw_event["room_id"] = room_id
                    tasks += self.dispatch_event(
                        self._try_deserialize(StateEvent, raw_event),
                        source=SyncStream.LEFT_ROOM | SyncStream.TIMELINE,
                    )
        return tasks

    def start(self, filter_data: FilterID | Filter | None) -> asyncio.Future:
        """
        Start syncing with the server. Can be stopped with :meth:`stop`.

        Args:
            filter_data: The filter data or filter ID to use for syncing.
        """
        if self.syncing_task is not None:
            self.syncing_task.cancel()
        self.syncing_task = asyncio.create_task(self._try_start(filter_data))
        return self.syncing_task

    async def _try_start(self, filter_data: FilterID | Filter | None) -> None:
        try:
            if isinstance(filter_data, Filter):
                filter_data = await self.create_filter(filter_data)
            await self._start(filter_data)
        except asyncio.CancelledError:
            self.log.debug("Syncing cancelled")
        except Exception as e:
            self.log.critical("Fatal error while syncing", exc_info=True)
            await self.run_internal_event(InternalEventType.SYNC_STOPPED, error=e)
            return
        except BaseException as e:
            self.log.warning(
                f"Syncing stopped with unexpected {e.__class__.__name__}", exc_info=True
            )
            raise
        else:
            self.log.debug("Syncing stopped without exception")
        await self.run_internal_event(InternalEventType.SYNC_STOPPED, error=None)

    async def _start(self, filter_id: FilterID | None) -> None:
        fail_sleep = 5
        is_first = True

        self.log.debug("Starting syncing")
        next_batch = await self.sync_store.get_next_batch()
        await self.run_internal_event(InternalEventType.SYNC_STARTED)
        timeout = 30
        while True:
            current_batch = next_batch
            start = time.monotonic()
            try:
                data = await self.sync(
                    since=current_batch,
                    filter_id=filter_id,
                    set_presence=self.presence,
                    timeout=timeout * 1000,
                )
            except (asyncio.CancelledError, MUnknownToken):
                raise
            except Exception as e:
                self.log.warning(
                    f"Sync request errored: {type(e).__name__}: {e}, waiting {fail_sleep}"
                    " seconds before continuing"
                )
                await self.run_internal_event(
                    InternalEventType.SYNC_ERRORED, error=e, sleep_for=fail_sleep
                )
                await asyncio.sleep(fail_sleep)
                if fail_sleep < 320:
                    fail_sleep *= 2
                continue
            if fail_sleep != 5:
                self.log.debug("Sync error resolved")
            fail_sleep = 5

            duration = time.monotonic() - start
            if current_batch and duration > timeout + 10:
                self.log.warning(f"Sync request ({current_batch}) took {duration:.3f} seconds")

            is_initial = not current_batch
            data["net.maunium.mautrix"] = {
                "is_initial": is_initial,
                "is_first": is_first,
            }
            next_batch = data.get("next_batch")
            try:
                await self.sync_store.put_next_batch(next_batch)
            except Exception:
                self.log.warning("Failed to store next batch", exc_info=True)
            await self.run_internal_event(InternalEventType.SYNC_SUCCESSFUL, data=data)
            if (self.ignore_first_sync and is_first) or (self.ignore_initial_sync and is_initial):
                is_first = False
                continue
            is_first = False
            self.log.silly(f"Starting sync handling ({current_batch})")
            start = time.monotonic()
            try:
                tasks = self.handle_sync(data)
                await asyncio.gather(*tasks)
            except Exception:
                self.log.exception(f"Sync handling ({current_batch}) errored")
            else:
                self.log.silly(f"Finished sync handling ({current_batch})")
            finally:
                duration = time.monotonic() - start
                if duration > 10:
                    self.log.warning(
                        f"Sync handling ({current_batch}) took {duration:.3f} seconds"
                    )

    def stop(self) -> None:
        """
        Stop a sync started with :meth:`start`.
        """
        if self.syncing_task:
            self.syncing_task.cancel()
            self.syncing_task = None

    @abstractmethod
    async def create_filter(self, filter_params: Filter) -> FilterID:
        pass

    @abstractmethod
    async def sync(
        self,
        since: SyncToken = None,
        timeout: int = 30000,
        filter_id: FilterID = None,
        full_state: bool = False,
        set_presence: PresenceState = None,
    ) -> JSON:
        pass



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
import asyncio

import olm

from mautrix.errors import DecryptionError, MatchingSessionDecryptionError
from mautrix.types import (
    DecryptedOlmEvent,
    EncryptedOlmEventContent,
    EncryptionAlgorithm,
    IdentityKey,
    OlmCiphertext,
    OlmMsgType,
    ToDeviceEvent,
    UserID,
)
from mautrix.util import background_task

from .base import BaseOlmMachine
from .sessions import Session


class OlmDecryptionMachine(BaseOlmMachine):
    async def _decrypt_olm_event(self, evt: ToDeviceEvent) -> DecryptedOlmEvent:
        if not isinstance(evt.content, EncryptedOlmEventContent):
            raise DecryptionError("unsupported event content class")
        elif evt.content.algorithm != EncryptionAlgorithm.OLM_V1:
            raise DecryptionError("unsupported event encryption algorithm")
        try:
            own_content = evt.content.ciphertext[self.account.identity_key]
        except KeyError:
            raise DecryptionError("olm event doesn't contain ciphertext for this device")

        self.log.debug(
            f"Decrypting to-device olm event from {evt.sender}/{evt.content.sender_key}"
        )
        plaintext = await self._decrypt_olm_ciphertext(
            evt.sender, evt.content.sender_key, own_content
        )

        try:
            decrypted_evt: DecryptedOlmEvent = DecryptedOlmEvent.parse_json(plaintext)
        except Exception:
            self.log.trace("Failed to parse olm event plaintext: %s", plaintext)
            raise
        if decrypted_evt.sender != evt.sender:
            raise DecryptionError("mismatched sender in olm payload")
        elif decrypted_evt.recipient != self.client.mxid:
            raise DecryptionError("mismatched recipient in olm payload")
        elif decrypted_evt.recipient_keys.ed25519 != self.account.signing_key:
            raise DecryptionError("mismatched recipient key in olm payload")
        decrypted_evt.sender_key = evt.content.sender_key
        decrypted_evt.source = evt
        self.log.debug(
            f"Successfully decrypted olm event from {evt.sender}/{decrypted_evt.sender_device} "
            f"(sender key: {decrypted_evt.sender_key} into a {decrypted_evt.type}"
        )
        return decrypted_evt

    async def _decrypt_olm_ciphertext(
        self, sender: UserID, sender_key: IdentityKey, message: OlmCiphertext
    ) -> str:
        if message.type not in (OlmMsgType.PREKEY, OlmMsgType.MESSAGE):
            raise DecryptionError("unsupported olm message type")

        try:
            plaintext = await self._try_decrypt_olm_ciphertext(sender_key, message)
        except MatchingSessionDecryptionError:
            self.log.warning(
                f"Found matching session yet decryption failed for sender {sender}"
                f" with key {sender_key}"
            )
            background_task.create(self._unwedge_session(sender, sender_key))
            raise

        if not plaintext:
            if message.type != OlmMsgType.PREKEY:
                background_task.create(self._unwedge_session(sender, sender_key))
                raise DecryptionError("Decryption failed for normal message")

            self.log.trace(f"Trying to create inbound session for {sender}/{sender_key}")
            try:
                session = await self._create_inbound_session(sender_key, message.body)
            except olm.OlmSessionError as e:
                background_task.create(self._unwedge_session(sender, sender_key))
                raise DecryptionError("Failed to create new session from prekey message") from e
            self.log.debug(
                f"Created inbound session {session.id} for {sender} (sender key: {sender_key})"
            )

            try:
                plaintext = session.decrypt(message)
            except olm.OlmSessionError as e:
                raise DecryptionError(
                    "Failed to decrypt olm event with session created from prekey message"
                ) from e

            await self.crypto_store.update_session(sender_key, session)

        return plaintext

    async def _try_decrypt_olm_ciphertext(
        self, sender_key: IdentityKey, message: OlmCiphertext
    ) -> Optional[str]:
        sessions = await self.crypto_store.get_sessions(sender_key)
        for session in sessions:
            if message.type == OlmMsgType.PREKEY and not session.matches(message.body):
                continue

            try:
                plaintext = session.decrypt(message)
            except olm.OlmSessionError as e:
                if message.type == OlmMsgType.PREKEY:
                    raise MatchingSessionDecryptionError(
                        "decryption failed with matching session"
                    ) from e
            else:
                await self.crypto_store.update_session(sender_key, session)
                return plaintext
        return None

    async def _create_inbound_session(self, sender_key: IdentityKey, ciphertext: str) -> Session:
        session = self.account.new_inbound_session(sender_key, ciphertext)
        await self.crypto_store.put_account(self.account)
        await self.crypto_store.add_session(sender_key, session)
        return session

    async def _unwedge_session(self, sender: UserID, sender_key: IdentityKey) -> None:
        raise NotImplementedError()



__version__ = "0.20.0rc1"
__author__ = "Tulir Asokan <tulir@maunium.net>"
__all__ = [
    "api",
    "appservice",
    "bridge",
    "client",
    "crypto",
    "errors",
    "util",
    "types",
    "__optional_imports__",
]

from typing import TYPE_CHECKING

__optional_imports__ = TYPE_CHECKING
